try:
    import typing  # help IDEs with type-hinting inside docstrings
except ImportError:
    pass
import numpy as np
import scipy.stats
import six

from .._base import FitEnsembleBase, FitEnsembleException
from ..tools.ensemble import EnsembleVariable, EnsembleVariablePlotter
from .cost import XYCostFunction_Chi2
from .fit import XYFit

import matplotlib as mpl
from matplotlib import pyplot as plt
from matplotlib import gridspec as gs


__all__ = ["XYFitEnsemble"]


def _heuristic_optimal_subplot_grid_size(n_subplots, aspect_ratio_priority=0.5):
    def f2(s, k):
        if n_subplots > s * (s + k):
            return 100000
        return ((s * (s + k) - n_subplots) ** 2 * (1.0 - aspect_ratio_priority)
                + (float(k) / float(s)) ** 2 * (aspect_ratio_priority))

    _optimal_f = np.inf
    _optimal_sk = n_subplots, 0
    for s in six.moves.range(1, n_subplots):
        for k in six.moves.range(0, n_subplots):
            _f = f2(s, k)
            if _f < _optimal_f:
                _optimal_f = _f
                _optimal_sk = s, k

    s, k = _optimal_sk
    return s, s+k


class XYFitEnsembleException(FitEnsembleException):
    pass


class XYFitEnsemble(FitEnsembleBase):
    """
    Object for generating ensembles of fits to *xy* pseudo-data generated according to the
    specified uncertainty model.

    After constructing an :py:obj:`~kafe2.fit.XYFitEnsemble` object, an error model should be added
    to it. This is done as for :py:obj:`~kafe2.fit.XYFit` objects by using the
    :py:meth:`~kafe2.fit.XYFitEnsemble.add_error` or :py:meth:`~kafe2.fit.XYFitEnsemble.add_matrix_error`
    methods.

    Once an uncertainty model is provided, the fit ensemble can be generated by using the
    :py:meth:`~kafe2.fit.XYFitEnsemble.run` method. This method starts by generating a pseudo-dataset in such a way
    that the empirical distribution of the data corresponds to the specified uncertainty model. It then
    fits the model to the pseudo-data and extracts information from the fit, such as the resulting parameter
    values or the value of the cost function at the minimum. This is repeated a large number of times
    in order to evaluate the whole ensemble in a statistically meaningful way.

    The ensemble result can be visualized by using the :py:meth:`~kafe2.fit.XYFitEnsemble.plot_results` method.

    .. TODO Expand section
    """
    FIT_TYPE = XYFit

    AVAILABLE_STATISTICS = {
        'mean': EnsembleVariable.mean,
        'mean_error': EnsembleVariable.mean_error,
        'std': EnsembleVariable.std,
        'skew': EnsembleVariable.skew,
        'kurtosis': EnsembleVariable.kurtosis,
        'cor_mat': EnsembleVariable.cor_mat,
        'cov_mat': EnsembleVariable.cov_mat,
    }
    _DEFAULT_STATISTICS = {'mean', 'std'}

    def __init__(self, n_experiments, x_support, model_function, model_parameters,
                 cost_function=XYCostFunction_Chi2(axes_to_use='y', errors_to_use='covariance'),
                 requested_results=None):
        """Construct an :py:obj:`~kafe2.fit.XYFitEnsemble` object.

        :param n_experiments: Number of pseudo experiments to perform.
        :type n_experiments: int
        :param x_support: *x* values to use as support for calculating the "true" model
            ("true" *x*).
        :type x_support: typing.Sequence[float]
        :param model_function: The model function. Either a
            :py:class:`~kafe2.fit.indexed.XYModelFunction` object or an unwrapped native Python
            function.
        :type model_function: typing.Callable
        :param model_parameters: Parameters of the "true" model
        :type model_parameters: typing.Sequence[float]
        :param cost_function: The cost function used for the fits. Either a
            :py:class:`~kafe2.fit._base.CostFunctionBase`-derived object or an unwrapped native
            Python function.
        :type cost_function: typing.Callable
        :param requested_results: List of result variables to collect for each toy fit. If
            :py:obj:`None` it will default to ``('y_pulls', 'parameter_pulls', 'cost')``.
        :type requested_results: typing.Sequence[str] or None.
        """
        self._n_exp = n_experiments
        self._ref_x_data = np.asarray(x_support, dtype=float)
        self._model_function = model_function
        self._model_parameters = np.asarray(model_parameters)
        self._cost_function = cost_function
        self._n_par = len(self._model_parameters)

        # initialize an `XYFit` object for performing the toy fits
        # need some dummy initial data values in order to initialize a Fit object
        self._ref_y_data = self._model_function(self._ref_x_data, *self._model_parameters)

        # initialize Fit object used for fitting the pseudo-data
        self._toy_fit = XYFit(xy_data=[self._ref_x_data, self._ref_y_data],
                              model_function=self._model_function,
                              cost_function=self._cost_function)

        # set the model parameters of the toy fit to the reference values
        self._set_toy_fit_parameters_to_reference()

        # get reference quantities (y data, covariance matrices...) from toy fit
        self._update_reference_quantities_from_toy_fit()

        # store and validate names of requested ensemble variables
        self._requested_results = requested_results
        if self._requested_results is None:
            self._requested_results = self._DEFAULT_RESULTS
        else:
            # validate list of results requested by user
            _unavailable_results = set(self._requested_results) - set(self.AVAILABLE_RESULTS.keys())
            if _unavailable_results:
                raise ValueError("Requested unavailable result variable(s): %r"
                                 % (_unavailable_results,))

        # initialize `EnsembleVariable` objects to store ensembles
        self._initialize_ensemble_variables()

    def _set_toy_fit_parameters_to_reference(self):
        """set the model parameters of the toy fit to the reference values"""
        self._toy_fit._param_model._model_parameters = self._model_parameters
        self._toy_fit._param_model._pm_calculation_stale = True

    def _generate_pseudodata(self):
        """generate new pseudo-data according to fit error model and commit to data container"""

        if not self._toy_fit.has_errors:
            raise FitEnsembleException("Cannot generate fit ensemble: no error model specified!")

        # -- generate 'x' data
        _x_data = self._ref_x_data.copy()

        if self._toy_fit.data_container.has_x_errors:
            # smear x data according to the total 'x' covariance matrix
            # TODO: only gaussian smearing is implemented -> more?
            _x_jitter = np.random.multivariate_normal(
                np.zeros_like(_x_data),
                self._ref_x_cov_mat)
            _x_data += _x_jitter

        _y_data = self._toy_fit.eval_model_function(x=_x_data,
                                                    model_parameters=self._model_parameters)

        # smear y data according to the total 'y' covariance matrix
        # TODO: only gaussian smearing is implemented -> more?
        _y_jitter = np.random.multivariate_normal(
            np.zeros_like(_y_data),
            self._ref_y_cov_mat)
        _y_data += _y_jitter

        # update toy fit data container
        self._toy_fit.data_container.x = _x_data
        self._toy_fit.data_container.y = _y_data

    def _gather_results_from_toy_fit(self, i_exp):
        for _var_name in self._requested_results:
            self._ensemble_variables[_var_name].set_value(index=i_exp, variable_value=self._get_var(_var_name))

    def _do_toy_fit(self):
        """run fit with current pseudo-data"""
        self._toy_fit.do_fit()

    def _get_var(self, var_name):
        """get the value of the result variables for the current fit"""
        return self.AVAILABLE_RESULTS[var_name].fget(self)

    def _initialize_ensemble_variables(self):
        self._ensemble_variables = {}
        self._ensemble_variable_plotters = {}
        if 'y_pulls' in self._requested_results:
            self._ensemble_variables['y_pulls'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp, self.n_dat)),
                distribution=scipy.stats.norm,
                distribution_parameters=dict(loc=0, scale=1)
            )
            self._ensemble_variable_plotters['y_pulls'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['y_pulls'],
                value_ranges=(-3, 3),
                variable_labels=['Pull $y_{%d}$' % (_i,) for _i in six.moves.range(1, self.n_dat+1)]
            )

        if 'x_data' in self._requested_results:
            self._ensemble_variables['x_data'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp, self.n_dat)),
                distribution=scipy.stats.norm,
                distribution_parameters=dict(loc=self._ref_x_data, scale=self._toy_fit.x_total_error)
            )
            self._ensemble_variable_plotters['x_data'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['x_data'],
                value_ranges=np.array([self._ref_x_data - 3 * self._toy_fit.x_total_error,
                                       self._ref_x_data + 3 * self._toy_fit.x_total_error]).T,
                variable_labels=['$x_{%d}$' % (_i,) for _i in six.moves.range(1, self.n_dat+1)]
            )

        if 'y_data' in self._requested_results:
            self._ensemble_variables['y_data'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp, self.n_dat)),
                distribution=scipy.stats.norm,
                distribution_parameters=dict(loc=self._ref_y_data, scale=self._ref_projected_xy_err)
            )
            self._ensemble_variable_plotters['y_data'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['y_data'],
                value_ranges=np.array([self._ref_y_data-3*self._ref_projected_xy_err,
                                       self._ref_y_data+3*self._ref_projected_xy_err]).T,
                variable_labels=['$y_{%d}$' % (_i,) for _i in six.moves.range(1, self.n_dat+1)]
            )

        if 'y_model' in self._requested_results:
            self._ensemble_variables['y_model'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp, self.n_dat)),
                distribution=scipy.stats.norm,
                distribution_parameters=dict(loc=self._ref_y_data, scale=self._ref_projected_xy_err)
            )
            self._ensemble_variable_plotters['y_model'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['y_model'],
                value_ranges=np.array([self._ref_y_data-3*self._ref_projected_xy_err,
                                       self._ref_y_data+3*self._ref_projected_xy_err]).T,
                variable_labels=['$f(x_{%d})$' % (_i,) for _i in six.moves.range(1, self.n_dat+1)]
            )

        if 'parameter_pulls' in self._requested_results:
            self._ensemble_variables['parameter_pulls'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp, self._n_par)),
                distribution=scipy.stats.norm,
                distribution_parameters=dict(loc=0, scale=1)
            )
            self._ensemble_variable_plotters['parameter_pulls'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['parameter_pulls'],
                value_ranges=(-3, 3),
                variable_labels=["Pull ${}$".format(_arg_formatter.latex_name)
                                 for _arg_formatter in self._toy_fit._model_function.formatter.arg_formatters]
            )

        if 'cost' in self._requested_results:
            self._ensemble_variables['cost'] = EnsembleVariable(
                ensemble_array=np.zeros((self._n_exp,)),
                distribution=scipy.stats.chi2,  # FIXME: assume chi2 for all cost functions -> change
                distribution_parameters=dict(loc=0, df=self.n_df)
            )
            self._ensemble_variable_plotters['cost'] = EnsembleVariablePlotter(
                ensemble_variable=self._ensemble_variables['cost'],
                value_ranges=(0, 3*self.n_df),
                variable_labels="${}$".format(self._toy_fit._cost_function.formatter.latex_name)
            )

    def _make_figure_gs(self, figsize=(8, 8), nrows=1, ncols=1,
                        left=0.1, bottom=0.1,
                        right=0.9, top=0.9):
        """create a new matplotlib figure with a GridSpec controlling the subplot layout"""
        _fig = plt.figure(figsize=figsize)  # defaults from matplotlibrc
        _gs = gs.GridSpec(nrows=nrows,
                          ncols=ncols,
                          left=left,
                          bottom=bottom,
                          right=right,
                          top=top,
                          wspace=None,
                          hspace=None,
                          height_ratios=None)
        return _fig, _gs

    def _update_reference_quantities_from_toy_fit(self):
        self._ref_y_data = self._toy_fit.eval_model_function(x=self._ref_x_data,
                                                             model_parameters=self._model_parameters)
        self._ref_x_cov_mat = self._toy_fit.x_total_cov_mat
        self._ref_y_cov_mat = self._toy_fit.y_total_cov_mat
        self._ref_projected_xy_cov_mat = self._toy_fit.total_cov_mat
        self._ref_x_err = self._toy_fit.x_total_error
        self._ref_y_err = self._toy_fit.y_total_error
        self._ref_projected_xy_err = self._toy_fit.total_error

    # -- private properties

    @property
    def _x_data(self):
        """property for ensemble variable 'x_data'"""
        return self._toy_fit.x

    @property
    def _parameter_pulls(self):
        """property for ensemble variable 'parameter_pulls'"""
        return (self._toy_fit.parameter_values - self._model_parameters)/self._toy_fit.parameter_errors

    @property
    def _y_data(self):
        """property for ensemble variable 'y_data'"""
        return self._toy_fit.y_data

    @property
    def _y_model(self):
        """property for ensemble variable 'y_model'"""
        return self._toy_fit.y_model

    @property
    def _y_pulls(self):
        """property for ensemble variable 'y_pulls'"""
        return (self._toy_fit.y_data - self._toy_fit.y_model) / self._toy_fit.y_total_error

    @property
    def _cost(self):
        """property for ensemble variable 'cost'"""
        return self._toy_fit.cost_function_value

    # -- public properties

    @property
    def n_exp(self):
        """The number of pseudo-experiments to perform.

        :rtype: int
        """
        return self._n_exp

    @property
    def n_par(self):
        """The number of parameters.

        :rtype: int
        """
        return self._n_par

    @property
    def n_dat(self):
        """The number of data points used for the fit.

        :rtype: int
        """
        return self._toy_fit.data_container.size

    @property
    def n_df(self):
        """The number of degrees of freedom for the fit

        :rtype: int
        """
        # FIXME: not generally true -> update to handle constrained parameters
        # TODO: not applicable for all cost functions -> find a flexible solution
        return self.n_dat - self.n_par

    # -- public methods

    def add_error(self, axis, err_val, name=None, correlation=0, relative=False, reference='data'):
        self._toy_fit.add_error(axis=axis, err_val=err_val,
                                name=name, correlation=correlation, relative=relative,
                                reference=reference)
        self._update_reference_quantities_from_toy_fit()  # recompute reference errors

    # "inherit" docstring
    add_error.__doc__ = XYFit.add_error.__doc__

    def add_matrix_error(self, axis, err_matrix, matrix_type, name=None, err_val=None, relative=False, reference='data'):
        self._toy_fit.add_matrix_error(axis=axis, err_matrix=err_matrix,
                                       matrix_type=matrix_type, name=name, err_val=err_val,
                                       relative=relative, reference=reference)
        self._update_reference_quantities_from_toy_fit()  # recompute reference errors

    # "inherit" docstring
    add_matrix_error.__doc__ = XYFit.add_matrix_error.__doc__

    def run(self):
        """Perform the pseudo-experiments. Retrieve and store the requested fit result variables."""
        self._set_toy_fit_parameters_to_reference()
        self._update_reference_quantities_from_toy_fit()
        self._initialize_ensemble_variables()
        for _i_exp in six.moves.range(self.n_exp):
            self._generate_pseudodata()
            self._do_toy_fit()
            self._gather_results_from_toy_fit(_i_exp)

    def get_results(self, *results):
        """Return a dictionary containing the ensembles of result variables.

        :param results: Names of result variables to retrieve. Calling without arguments retrieves
            *all* collected results.
        :type results: typing.Iterable[str]
        :rtype: dict
        """
        if not results:
            results = self._requested_results
        else:
            # validate list of results requested by user
            _unavailable_results = set(self._requested_results) - set(self.AVAILABLE_RESULTS.keys())
            if _unavailable_results:
                raise ValueError("Requested unavailable result variable(s): %r"
                                 % (_unavailable_results,))

        _dict_to_return = dict()
        for _result_name in results:
            _var = self._ensemble_variables.get(_result_name, None)
            if _var is None:
                raise FitEnsembleException("Cannot retrieve result '{}': "
                                           "variable not collected!".format(_result_name))
            _dict_to_return[_result_name] = _var.values

        return _dict_to_return

    def get_results_statistics(self, results='all', statistics='all'):
        """Return a dictionary containing statistics (e.g. mean) of the result variables.

        :param results: Names of retrieved fit variable for which to return statistics. If
            ``'all'``, get statistics for all retrieved variables
        :type results: typing.Iterable[str] or str
        :param statistics: Names of statistics to retrieve for each result variable. If ``'all'``,
            get all statistics for each retrieved variable
        :type statistics: typing.Iterable[str] or str
        :rtype: dict
        """
        if results == 'all':
            results = self._requested_results

        if statistics == 'all':
            statistics = self.__class__._DEFAULT_STATISTICS

        _dict_to_return = dict()
        for _result_name in results:
            #_result_array = self._result_array_dicts.get(_result_name, None)
            _result_variable = self._ensemble_variables.get(_result_name, None)
            if _result_variable is None:
                raise FitEnsembleException("Cannot retrieve statistics for result "
                                           "variable '{}': variable not collected!".format(_result_name))

            _current_result_dict = _dict_to_return[_result_name] = dict()

            # calculate and store statistics
            for _stat_name in statistics:
                _stat_unbound_method = self.__class__.AVAILABLE_STATISTICS.get(_stat_name, None)
                if _stat_unbound_method is None:
                    raise FitEnsembleException(
                        "Unknown statistic '%s' requested!" % (_stat_name,))
                _stat = _stat_unbound_method.__get__(_result_variable, EnsembleVariable)
                _current_result_dict[_stat_name] = _stat

        return _dict_to_return

    def plot_result_distributions(self, results='all',
                                  show_legend=True):
        """Make plots with histograms of the requested fit variable values across all
        pseudo-experiments.

        :param results: Names of retrieved fit variable for which to generate plots. If
            ``'all'``, plots for all retrieved variables will be made.
        :type results: typing.Iterable[str] or str
        :param show_legend: If a legend is shown on each figure.
        :type show_legend: bool
        """
        if results == 'all':
            results = self._requested_results
        else:
            # validate list of results requested by user
            _unavailable_results = set(self._requested_results) - set(self.AVAILABLE_RESULTS.keys())
            if _unavailable_results:
                raise ValueError("Requested unavailable result variable(s): %r"
                                 % (_unavailable_results,))

        for _result_name in results:
            _result_variable = self._ensemble_variables.get(_result_name, None)

            if _result_variable is None:
                raise FitEnsembleException("Cannot plot result for variable '%s': "
                                           "variable not collected!" % (_result_name,))

            _result_variable_plotter = self._ensemble_variable_plotters.get(_result_name, None)

            if _result_variable_plotter is None:
                raise FitEnsembleException("Cannot plot result for variable '%s': "
                                           "no plotter defined!" % (_result_name,))

            # -- decide how to lay out plots depending on the result variable dimensionality
            if _result_variable.ndim == 0:
                # if the ensemble variable is a scalar,
                # plot it into a single `Axes` object
                _fig, _gs = self._make_figure_gs(figsize=(8, 8), nrows=1, ncols=1)
                _ax = plt.subplot(_gs[0, 0])
                # call the plotting routine on the axes grid
                _plot_result_dict = _result_variable_plotter.plot_hist(_ax)

            elif _result_variable.ndim == 1:
                # if the ensemble variable is a one-dimensional vector,
                # plot each entry into a separate `Axes` object and display
                # them in a grid-like layout
                _nplots = int(_result_variable.shape[0])
                _nrows, _ncols = _heuristic_optimal_subplot_grid_size(_nplots, aspect_ratio_priority=0.8)
                _fig, _gs = self._make_figure_gs(figsize=(8, 8), nrows=_nrows, ncols=_ncols)

                # create an array 'a' with a[i, j] = [i, j]
                _axes_grid = np.dstack((np.meshgrid(np.arange(_nrows), np.arange(_ncols))))
                # replace [i, j] by the `Axes` object for _gs[i, j] -> array of `Axes`
                _axes_grid = np.apply_along_axis(
                    lambda irow_icol: plt.subplot(_gs[irow_icol[0], irow_icol[1]]) if irow_icol[0]*_ncols+irow_icol[1] < _nplots else None,
                    -1, _axes_grid)
                # reshape the `Axes` array to match the variable shape
                _axes_grid = _axes_grid.T.flatten()[:_result_variable.shape[0]]
                # call the plotting routine on the axes grid
                _plot_result_dict = _result_variable_plotter.plot_hist(_axes_grid)

            elif _result_variable.ndim == 2:
                # if the ensemble variable is a two-dimensional vector,
                # plot the (i,j)-th entry into a an `Axes` object at the
                # (i,j)-th position in a grid

                _nrows = _result_variable.shape[0]
                _ncols = _result_variable.shape[1]

                _fig, _gs = self._make_figure_gs(figsize=(8, 8), nrows=_nrows, ncols=_ncols)

                # create an array 'a' with a[i, j] = [i, j]
                _axes_grid = np.dstack(reversed(np.meshgrid(np.arange(_nrows), np.arange(_ncols))))
                # replace [i, j] by the `Axes` object for _gs[i, j] -> array of `Axes`
                _axes_grid = np.apply_along_axis(
                    lambda irow_icol: plt.subplot(_gs[irow_icol[0], irow_icol[1]]),
                    -1, _axes_grid)
                # do not reshape _axes_grid -> its shape already matches variable shape

                # call the plotting routine on the axes grid
                _plot_result_dict = _result_variable_plotter.plot_hist(_axes_grid)
            else:
                # cannot plot variables with 3 or more dimensions...
                raise FitEnsembleException("Cannot plot result for variable '%s': variable entry dimensionality "
                                           "too high (%d)!" % (_result_name, _result_variable.ndim))

            if show_legend:
                _fig.legend(_plot_result_dict['legend_handles'],
                            _plot_result_dict['legend_labels'], loc='lower center')
                # add extra space at figure bottom for legend
                _figure_extra_bottom = 0.05 * len(_plot_result_dict['legend_labels'])
            else:
                # no extra space at figure bottom
                _figure_extra_bottom = 0.0

            _fig.canvas.set_window_title(_result_name)

            _gs.tight_layout(_fig,
                             pad=0.0, w_pad=0, h_pad=-0.2,
                             rect=(0.01, 0.02+_figure_extra_bottom, 0.98, 0.98))

        return _plot_result_dict

    def plot_result_scatter(self, results='all',
                                  show_legend=True):
        """Make scatter plots of the requested fit variable values across all pseudo-experiments.

        :param results: Names of retrieved fit variable for which to generate plots. If
            ``'all'``, plots for all retrieved variables will be made.
        :type results: typing.Iterable[str] or str
        :param show_legend: If a legend is shown on each figure.
        :type show_legend: bool
        """
        if results == 'all':
            results = self._requested_results
        else:
            # validate list of results requested by user
            _unavailable_results = set(self._requested_results) - set(self.AVAILABLE_RESULTS.keys())
            if _unavailable_results:
                raise ValueError("Requested unavailable result variable(s): %r"
                                 % (_unavailable_results,))

        for _result_name in results:
            _result_variable = self._ensemble_variables.get(_result_name, None)

            if _result_variable is None:
                raise FitEnsembleException("Cannot plot result for variable '%s': "
                                           "variable not collected!" % (_result_name,))

            _result_variable_plotter = self._ensemble_variable_plotters.get(_result_name, None)

            if _result_variable_plotter is None:
                raise FitEnsembleException("Cannot plot result for variable '%s': "
                                           "no plotter defined!" % (_result_name,))

            # -- decide how to lay out plots depending on the result variable dimensionality
            if _result_variable.ndim != 1:
                raise ValueError()

            # if the ensemble variable is a one-dimensional vector,
            # plot each entry into a separate `Axes` object and display
            # them in a grid-like layout
            _nrows = _ncols = int(_result_variable.shape[0])
            if _nrows <= 1:
                raise FitEnsembleException("Cannot create scatter plot for result variable '%s': "
                                           "vector has less than two entries!" % (_result_name,))
            _fig, _gs = self._make_figure_gs(figsize=(8, 8), nrows=_nrows-1, ncols=_ncols-1)

            # create an array 'a' with a[i, j] = [i, j]
            _axes_grid = np.dstack((np.meshgrid(np.arange(_nrows), np.arange(_ncols))))
            # replace [i, j] by the `Axes` object for _gs[i, j] -> array of `Axes`
            _axes_grid = np.apply_along_axis(
                lambda irow_icol: plt.subplot(_gs[irow_icol[0] - 1, irow_icol[1]]) if irow_icol[0] > irow_icol[1] else None,
                -1, _axes_grid)

            # call the plotting routine on the axes grid
            _plot_result_dict = _result_variable_plotter.plot_scatter(_axes_grid)

            if show_legend:
                _fig.legend(_plot_result_dict['legend_handles'],
                            _plot_result_dict['legend_labels'], loc='lower center')
                # add extra space at figure bottom for legend
                _figure_extra_bottom = 0.05 * len(_plot_result_dict['legend_labels'])
            else:
                # no extra space at figure bottom
                _figure_extra_bottom = 0.0

            _fig.canvas.set_window_title(_result_name)

            _gs.tight_layout(_fig,
                             pad=0.0, w_pad=0, h_pad=-0.2,
                             rect=(0.01, 0.02+_figure_extra_bottom, 0.98, 0.98))

        return _plot_result_dict


    AVAILABLE_RESULTS = {
        'parameter_pulls': _parameter_pulls,
        'x_data': _x_data,
        'y_pulls': _y_pulls,
        'cost': _cost,
        'y_data': _y_data,
        'y_model': _y_model,
    }
    _DEFAULT_RESULTS = {'y_pulls', 'parameter_pulls', 'cost'}



from __future__ import print_function

import collections
import matplotlib as mpl
import numpy as np
import scipy.stats
import six

from ...core.error import CovMat


__all__ = ['EnsembleVariable', 'EnsembleVariableProbabilityDistribution', 'EnsembleVariablePlotter']


def cycle_axes(array, k):
    """
    Cycle the dimensions of an array forward by `k` (such that the
    the `i`-th axis becomes the `i+k`-th axis, modulo the array
    dimensionality).

    :param array: array for which to cycle axes
    :type array: ``numpy.ndarray``
    :param k: number of positions to cycle axes.
    :type k: int with ``abs(n) < array.ndim``
    :return: the array with moved axes
    :rtype: ``numpy.ndarray``
    """
    _nd = array.ndim
    k = k % _nd
    _axes_permutation = (np.arange(_nd) - k) % _nd
    return np.transpose(array, _axes_permutation)


def expand_to_ndim(array, target_ndim, direction='right'):
    """Add axes to an array until the desired number of dimensions is reached.

    :param array: array to expand
    :type array: ``numpy.ndarray``
    :param target_ndim: number of dimensions to reach
    :type target_ndim: int
    :param direction: direction of expansion
    :type direction: 'left' or 'right'
    :return: expanded array
    :rtype: ``numpy.ndarray``
    """
    _new_array = array
    if direction == 'right':
        _append_pos = -1
    elif direction == 'left':
        _append_pos = 0
    else:
        raise ValueError("Unknown direction specification '{}': "
                         "expected 'left' or 'right'!".format(direction))
    while _new_array.ndim < target_ndim:
        _new_array = np.expand_dims(_new_array, _append_pos)
    return _new_array

def broadcast_to_shape(array, shape, scheme='default'):
    """
    Broadcast a ``numpy.ndarray`` to a shape according to different schemes.

    :param array: the array to broadcast
    :type array: ``numpy.ndarray``
    :param shape: the shape against which to broadcast the array
    :type shape: tuple of ``int``
    :param scheme: the name of the scheme
    :type scheme: 'default', 'transposed', 'expand_left' or 'expand_right'

    :return: the broadcasted array
    """
    if scheme == 'default':
        # default numpy broadcasting scheme:
        #        shape:  a x b x c
        #        array:      b x 1
        #      ---------------------
        #       result:  a x b x c
        return np.broadcast_to(array, shape=shape)
    if scheme == 'expand_right_successive':
        # dimensions are added to the array (on the right) until
        # a broadcastable situation is reached
        # target shape to the array shape:
        #        shape:      a x b
        #        array:  n x a
        #      ---------------------
        #       result:  n x a x 1
        _new_array = array
        while True:
            _can_bcast = all([_i == _j or _i == 1 or _j == 1 for _i, _j in zip(reversed(_new_array.shape), reversed(shape))])
            if _can_bcast:
                break
            _new_array = np.expand_dims(_new_array, -1)
        return _new_array
    if scheme == 'expand_left_successive':
        # dimensions are added to the array (on the left) until
        # a broadcastable situation is reached
        # target shape to the array shape:
        #        shape:  a x b
        #        array:      b x n
        #      ---------------------
        #       result:  1 x b x n
        _new_array = array
        while True:
            _can_bcast = all([_i == _j or _i == 1 or _j == 1 for _i, _j in zip(_new_array.shape, shape)])
            if _can_bcast:
                break
            _new_array = np.expand_dims(_new_array, 0)
        return _new_array
    if scheme == 'transposed':
        # "transposed" numpy broadcasting scheme:
        # (broadcasting is done on the transposed arrays
        #  and the result is transposed back)
        #        shape:  a x b x c
        #        array:  a x 1
        #      ---------------------
        #       result:  a x b x c
        return np.broadcast_to(array.T, shape=reversed(shape)).T
    if scheme == 'expand_right':
        # `len(shape)` axes are added to the array shape (to the right)
        # and then broadcasting is done normally
        #        shape:          a x b x c  (self.ndim = 3)
        #        array:  n x m x 1 x 1 x 1
        #      -----------------------------
        #       result:  n x m x a x b x c
        # _new_shape = tuple(len(self._shape) * [1] + list(reversed(array.shape)))
        _new_shape = tuple(list(reversed(shape)) + list(reversed(array.shape)))
        return np.broadcast_to(array.T, _new_shape).T
    if scheme == 'expand_left':
        # `len(shape)` axes are added to the array shape (to the left)
        # and then the 'transposed' broadcasting is applied
        #        shape:  a x b x c          (self.ndim = 3)
        #        array:  1 x 1 x 1 x n x m
        #      -----------------------------
        #       result:  a x b x c x n x m
        _new_shape = tuple(list(shape) + list(array.shape))
        return np.broadcast_to(array, _new_shape)
    raise EnsembleError("Cannot broadcast array to ensemble shape: unknown scheme '{}'! "
                        "Available: {}".format(scheme, ('default', 'transposed', 'expand_right', 'expand_left',
                                                        'expand_right_successive', 'expand_left_successive')))


class EnsembleError(Exception):
    pass


class EnsembleVariable(object):
    """
    Object for storing a finite sample of realizations of a single (possibly multidimensional) random variable,
    forming a statistical ensemble.
    """

    def __init__(self, ensemble_array, dtype=float,
                 distribution=None, distribution_parameters=None):
        """
        Create an ensemble of realizations of random variables.

        :param ensemble_array: a statistical ensemble containing all realizations of the random variable.
                               **Note**: the size of the first axis is taken to be the sample size. Any
                               remaining array axes are understood to be part of the (multidimensional)
                               random variable itself.
        :type ensemble_array: `numpy.ndarray`
        :param sample_size: the size of the ensemble (number of realizations of the variable)
        :type sample_size: int
        :param variable_shape: the ndarray shape for *one* realization of the random variable
        :typle variable_shape: tuple of int (pass an empty `tuple()` or list `[]` for a scalar variable)
        :param dtype: underlying dtype of ensemble variable
        :type dtype: type
        :param distribution: probability distribution of the variable (e.g. `scipy.stats.norm`)
        :type distribution: a probability distribution from `scipy.stats` or ``None``
        :param distribution_parameters:  mapping of distribution parameter names to values. If values are
                                         sequences/arrays, the shape must match the `variable_shape`
        :type distribution_parameters: dict or ``None``
        """
        self._array = np.asarray(ensemble_array)
        self._total_shape = self._array.shape
        self._size = self._total_shape[0]
        self._shape = tuple()
        if self._array.ndim > 1:
            self._shape = self._total_shape[1:]

        self._dist = None
        if distribution is not None:
            self._dist = EnsembleVariableProbabilityDistribution(
                distribution=distribution,
                parameters=distribution_parameters,
                variable_shape=self._shape
            )

    @property
    def size(self):
        """The size of the sample."""
        return self._size

    @property
    def shape(self):
        """The shape of the random variable."""
        return self._shape

    @property
    def ndim(self):
        """The dimensionality of the random variable."""
        return self._array.ndim - 1  # do not include the sample size dimension

    @property
    def values(self):
        """A (possibly) multidimensional array containing all realization of the ensemble variable."""
        return self._array

    @property
    def mean(self):
        """The mean of the ensemble variable across all realizations."""
        return np.mean(self._array, axis=0)

    @property
    def mean_error(self):
        """The standard error of the mean -> standard deviation/sqrt(N)"""
        return self.std / np.sqrt(self.size)

    @property
    def std(self):
        """The standard deviation of the ensemble variable across all realizations."""
        return np.std(self._array, axis=0)

    @property
    def skew(self):
        """The skew of the ensemble variable across all realizations."""
        return scipy.stats.skew(self._array, axis=0)

    @property
    def kurtosis(self):
        """The kurtosis of the ensemble variable across all realizations."""
        return scipy.stats.kurtosis(self._array, axis=0)

    @property
    def cov_mat(self):
        """The sample covariance matrix (only available for one-dimensional ensemble variables)."""
        if self.ndim == 0:
            # trivial covariance matrix
            return np.array([[self.std**2]])
        if self.ndim == 1:
            return np.cov(self._array.T)

        raise EnsembleError("Cannot calculate covariance matrix: ensemble variable must "
                            "have dimension 1 (got {})".format(self.ndim))

    @property
    def cor_mat(self):
        """The sample correlation matrix (only available for one-dimensional ensemble variables)."""
        if self.ndim == 0:
            # trivial correlation matrix
            return np.array([[1.0]])
        if self.ndim == 1:
            return CovMat(self.cov_mat).cor_mat

        raise EnsembleError("Cannot calculate correlation matrix: ensemble variable must "
                            "have dimension 1 (got {})".format(self.ndim))

    @property
    def dist(self):
        """An object representing the (expected) probability distribution of the variable."""
        return self._dist  # can be ``None``

    def set_value(self, index, variable_value):
        """Set the value of the `index`-th realization of the ensemble variable."""
        # TODO (?) validate index and/or value shape?
        try:
            self._array[index, :] = variable_value
        except IndexError:
            # for scalar ensemble variables
            self._array[index] = variable_value


class EnsembleVariableProbabilityDistribution(object):
    """
    Object for storing the probability distribution for an ensemble variable.
    """

    def __init__(self, distribution, parameters, variable_shape):
        """
        Create a probability distribution object for a random variable.

        :param distribution: a ``scipy.stats`` probability distribution
        :type distribution: e.g. ``scipy.stats.norm``
        :param parameters: the parameters of the probability distribution
        :type parameters: ``dict`` mapping parameter names to values (or value arrays)
        :param variable_shape: the shape of the "``numpy.ndarray``-valued" random variable
        :type variable_shape: the shape of the "``numpy.ndarray``-valued" random variable
        """
        # self._dist_func = distribution
        self._shape = variable_shape

        self._dist_param_values_dict = {}
        for _dist_par_name, _dist_par_value in six.iteritems(parameters):
            # wrap lists and/or tuples in numpy.ndarray
            if isinstance(_dist_par_value, collections.Sequence) and not isinstance(_dist_par_value,
                                                                                    six.string_types[0]):
                _dist_par_value = np.array(_dist_par_value)

            if isinstance(_dist_par_value, np.ndarray):
                if _dist_par_value.ndim == 0:
                    _dist_par_value = np.ones(self._shape) * _dist_par_value
                if _dist_par_value.shape != self._shape:
                    raise EnsembleError("Distribution parameter '{}' must be a scalar or have shape "
                                        "{}: got {}".format(_dist_par_name, self._shape, _dist_par_value.shape))
            else:
                _dist_par_value = np.ones(self._shape) * _dist_par_value
            self._dist_param_values_dict[_dist_par_name] = _dist_par_value

        self._dist_func = distribution(**self._dist_param_values_dict)

    @property
    def ndim(self):
        """The dimensionality of the random variable."""
        return len(self._shape)  # do not include the sample size dimension

    @property
    def mean(self):
        """The expectation value of the probability distribution"""
        return self._dist_func.mean()

    @property
    def var(self):
        """The variance of the probability distribution."""
        return self._dist_func.var()

    @property
    def std(self):
        """The standard deviation (square root of the variance) of the probability distribution."""
        return self._dist_func.std()

    @property
    def skew(self):
        """The skewness (third standardized moment) of the probability distribution."""
        return self.standardized_moment(3)

    @property
    def kurtosis(self):
        """The kurtosis (fourth standardized moment) of the probability distribution."""
        return self.standardized_moment(4)

    def standardized_moment(self, n):
        """The n-th standardized moment of the distribution"""
        if self.ndim != 0:
            raise NotImplementedError("Standardized moment calculation not available for non-scalar variables.")
        return self._dist_func.expect(lambda x: ((x - self.mean) / self.std) ** n)

    def moment(self, n):
        """The n-th moment of the distribution"""
        if self.ndim != 0:
            raise NotImplementedError("Moment calculation not available for non-scalar variables.")
        return self._dist_func.moment(n)

    def eval(self, x, x_contains_var_shape=False):
        """
        Evaluate the probability distribution/mass at a given point/an array of given points.

        If `x` is a scalar, the probability distribution/mass function
        will be evaluated at the single point `x` for *all*
        different parameter values specified when constructing the object.
        If the parameter values are specified as *p*-dimensional
        arrays, the resulting array will also have *p* dimensions,
        and the shape will match that of the parameter arrays. For example::

          par_shape =    (a, b, c)  # ndim = 3 (= p)
          x.shape =      ()         # ndim = 0 (scalar, n = 0)
          result.shape = (a, b, c)  # ndim = 3

        If `x` is an *n*-dimensional array with shape `x.shape`,
        the probability distribution/mass function
        will be evaluated at every value of `x`.
        In this case, the resulting array will have *p+n* dimensions,
        with the first *p* corresponding to the parameter arrays
        and the last *n*  corresponding to `x.shape`. For example::

          par_shape =    (a, b, c)        # ndim = 3 (= p)
          x.shape =      (m, n)           # ndim = 2 (= n)
          result.shape = (a, b, c, m, n)  # ndim = p+n = 5

        Alternatively, if the flag ``x_contains_var_shape`` is set, the
        first dimensions of ``x`` are assumed to have the
        same shape as the parameter arrays. If this is the case, no
        broadcasting of ``x`` to the parameter arrays is done
        and the resulting array will have the same dimension and shape as
        the input array ``x``. For example::

          par_shape =    (a, b, c)        # ndim = 3 (= p)
          x.shape =      (a, b, c, m, n)  # ndim = p + 2 = 5
          result.shape = (a, b, c, m, n)  # ndim = p + 2 = 5

        :param x: value (array of values) at which to evaluate the pdf/pmf
        :type x: `numpy.ndarray`. If the alternative method above is used, the *first* entries of `x.shape`
                 must match the ensemble variable's shape
        :param x_contains_var_shape: if ``True``, the alternative method described above is used.
        :type x_contains_var_shape: bool
        :return:
        """
        # copy x so it is not modified
        x = np.asarray(x).copy()

        # if not already done, the `x` array shape must be expanded
        # so that the first dimensions match `self._shape`
        # -> transform x.shape from (m, n) to (a, b, c, m, n) -> 'expand_left'
        if not x_contains_var_shape:
            # reminder example:
            # par_shape =    (a, b, c)        # ndim = 3 (= p)
            # x.shape =      (m, n)           # ndim = 2 (= n)
            # new x.shape =  (a, b, c, m, n)  # ndim = p+n = 5
            x = broadcast_to_shape(x, shape=self._shape, scheme='expand_left')

        # before evaluating, `x` must have a shape that can be broadcast
        # together with the parameter array with shape = (a, b, c)
        # -> transform x.shape from (a, b, c, m, n) to (m, n, a, b, c)
        x = cycle_axes(x, -self.ndim)

        # evaluate pdf or pmf for x
        _y = None
        if _y is None:
            try:
                _y = self._dist_func.pdf(x)
            except AttributeError:
                pass
        if _y is None:
            try:
                _y = self._dist_func.pmf(x)
            except AttributeError:
                pass

        # after the evaluation, the result must be broadcast back to
        # the desired output shape
        # -> transform x.shape from (m, n, a, b, c) back to (a, b, c, m, n)
        return cycle_axes(_y, self.ndim)


class EnsembleVariablePlotter(object):
    """
    Object for storing the relevant properties and methods for plotting an ensemble variable.
    """

    _DEFAULT_PLOT_PDF_KWARGS = dict(marker='')
    #TODO fix Exception on import
    try:
        _DEFAULT_PLOT_EXPECTED_MEAN_KWARGS = dict(linewidth=1, marker='', linestyle='--',
                                              # use second color in default color cycle
                                              color=mpl.rcParams['axes.prop_cycle'].by_key()['color'][1])
    except:
        _DEFAULT_PLOT_EXPECTED_MEAN_KWARGS = None
    _DEFAULT_PLOT_OBSERVED_MEAN_KWARGS = dict(linewidth=1, marker='', color='k')
    _DEFAULT_PLOT_ONE_SIGMA_BAND_MEAN_KWARGS = dict(color='k', alpha=0.1)

    def __init__(self, ensemble_variable,
                 value_ranges,
                 variable_labels=None,
                 ensemble_label=None):
        """
        Create an object to handle the graphical representation of an ensemble variable.

        :param ensemble_variable: the ensemble variable to be plotted
        :type ensemble_variable: `~kafe2.fit.tools.ensemble.EnsembleVariable`
        :param value_ranges: the ...
        :type value_ranges: `numpy.ndarray` (shape must match the ensemble variable shape)
        :param axis_labels: the labels to appear on the axes
        :type axis_labels: `numpy.ndarray` (shape must match the ensemble variable shape)
        """
        self._var = ensemble_variable

        # check variable dimensionality (can only plot scalar, 1D and 2D)
        if self._var.ndim >= 3:
            raise EnsembleError("Cannot create plotter for ensemble variable: "
                                "dimensionality too high ({}>2)!".format(self._var.ndim))

        self._ensemble_label = ensemble_label or "{} pseudoexperiments".format(self._var.size)

        self._value_ranges = np.array(value_ranges)
        if self._value_ranges.shape[:-1] != self._var.shape:
            # prefix dimensions do not match -> expand array
            self._value_ranges = broadcast_to_shape(self._value_ranges, self._var.shape, scheme="expand_left")

        # expand, if needed, to at least 3 axes (needed for nested for loop below)
        self._value_ranges = expand_to_ndim(self._value_ranges, target_ndim=3, direction='left')

        self._variable_labels = np.array(variable_labels)
        if self._variable_labels is not None:
            self._variable_labels = broadcast_to_shape(self._variable_labels, self._var.shape, scheme="expand_left")
            self._variable_labels = expand_to_ndim(self._variable_labels, target_ndim=2, direction='left')

    def plot_hist(self, axes_array, show_y_ticks=False):
        """
        Plot each component of the (multidimensional) ensemble variable as a histogram.



        :param axes_array: an array of ``matplotlib`` axes objects in which to perform the plotting. The shape
                           of the axes array must match that of the ensemble variable. For scalar variables,
                           a single ``Axes`` can be provided.
        :type axes_array: ``numpy.ndarray`` of ``matplotlib`` ``Axes`` objects.

        :return: mapping containing plot metadata and other information
        :rtype: `dict`
        """
        axes_array = np.asarray(axes_array)

        if axes_array.shape != self._var.shape:
            raise EnsembleError("Shape of `axes_array` {} does not "
                                "match ensemble variable shape {}!".format(axes_array.shape, self._var.shape))

        _expected_means = None
        _expected_mean_errors = None
        _pdf_eval_ymax = None

        _nbins = 51  # TODO: config
        _pdf_eval_npts = 101  # TODO: config

        # get the expected mean and its error from the ensemble variable distribution
        _dist = self._var.dist
        if _dist is not None:
            _expected_means = np.atleast_2d(self._var.dist.mean)
            # TODO: presumably only valid/relevant for Gaussian -> solution for non-Gaussian?
            _expected_mean_errors = np.atleast_2d(self._var.dist.std)/np.sqrt(self._var.size)

            # evaluate PDF at pre-computed plot points
            _pdf_eval_x = np.apply_along_axis(
                lambda xminmax: np.linspace(xminmax[0], xminmax[1], _pdf_eval_npts), -1, np.squeeze(self._value_ranges))
            _pdf_eval_y = _dist.eval(_pdf_eval_x,
                                     x_contains_var_shape=True)

            # pad the evaluated PDF coordinates to at least 3 dimensions
            _pdf_eval_x = expand_to_ndim(_pdf_eval_x, 3, direction='left')
            _pdf_eval_y = expand_to_ndim(_pdf_eval_y, 3, direction='left')

            _pdf_eval_ymax = np.max(_pdf_eval_y, axis=-1)

        # get the observed mean (pad to at least 2 dimensions: one scalar per plot in 2D matrix)
        _observed_means = np.atleast_2d(self._var.mean)

        # get value array (pad to at least 3 dimensions: one 1D-array per plot in 2D matrix)
        _var_values = expand_to_ndim(self._var.values, 3, direction='right')

        _all_legend_handles = []
        _all_legend_labels = []

        _plot_result_dict = dict()
        for _index1, _axes in enumerate(np.atleast_2d(axes_array)):
            for _index2, _ax in enumerate(_axes):
                _data = _var_values[:, _index2, _index1]
                assert len(_data) == self._var.size

                _bin_contents, _bin_edges, _ = _ax.hist(
                    _data,
                    bins=_nbins,
                    range=self._value_ranges[_index1, _index2], # TODO: what about underflow/overflow?
                    label=self._ensemble_label
                )

                if _expected_means is not None:
                    # only show observed mean if expected mean is available
                    _ax.annotate(r"$\mu={}$".format(round(_observed_means[_index1, _index2], 2)),
                                 xycoords='data',
                                 xy=(_observed_means[_index1, _index2], 0),
                                 textcoords='offset points',
                                 xytext=(0, 25),
                                 fontsize=12,
                                 horizontalalignment='center',
                                 verticalalignment='bottom',
                                 arrowprops=dict(facecolor='k', shrink=.0)
                                 )
                    if _expected_mean_errors is not None and _expected_mean_errors[_index1, _index2]:
                        _mean_pull = (_observed_means[_index1, _index2] - _expected_means[_index1, _index2]) / _expected_mean_errors[_index1, _index2]
                        _ax.annotate(r"$({:+.2f}\sigma)$".format(round(_mean_pull, 2)),
                                     xycoords='data',
                                     xy=(_observed_means[_index1, _index2], 0),
                                     textcoords='offset points',
                                     xytext=(0, 40),
                                     fontsize=12,
                                     horizontalalignment='center',
                                     verticalalignment='bottom',
                                     arrowprops=dict(arrowstyle='-')
                                     )

                if self._value_ranges is not None:
                    _ax.set_xlim(self._value_ranges[_index1, _index2])

                if _dist is not None:
                    _pdf_label = "expected density"
                    # calculate normalization
                    _mean_bin_width = np.mean(_bin_edges[1:] - _bin_edges[:-1])
                    _n_entries = np.sum(_bin_contents)
                    _plot_prob_density_scale = _mean_bin_width * _n_entries
                    _pdf_x = _pdf_eval_x[_index1, _index2]
                    _pdf_y = _pdf_eval_y[_index1, _index2] * _plot_prob_density_scale

                    _ax.plot(_pdf_x, _pdf_y, label=_pdf_label, **self._DEFAULT_PLOT_PDF_KWARGS)

                if _expected_means is not None:
                    _ax.axvline(_expected_means[_index1, _index2], label="expected mean",
                                **self._DEFAULT_PLOT_EXPECTED_MEAN_KWARGS)
                    if _expected_mean_errors is not None:
                        _ax.axvspan(_expected_means[_index1, _index2] - _expected_mean_errors[_index1, _index2],
                                    _expected_means[_index1, _index2] + _expected_mean_errors[_index1, _index2],
                                    label="standard error of the mean",
                                    **self._DEFAULT_PLOT_ONE_SIGMA_BAND_MEAN_KWARGS)

                # ensure density appears with the same scaling across all axes
                if _pdf_eval_ymax is not None:
                    _ax.set_ylim((0, _pdf_eval_ymax[_index1, _index2] * _plot_prob_density_scale * 1.2))

                # set the x axis label
                if self._variable_labels is not None:
                    _xlabel = self._variable_labels[_index1, _index2]
                    if _xlabel is not None:
                        _ax.set_xlabel(_xlabel)

                # disable the y axis ticks
                if not show_y_ticks:
                    _ax.yaxis.set_ticks([])

                # collect legend handles and labels
                _hs, _ls = _ax.get_legend_handles_labels()
                _all_legend_handles += tuple(_hs)
                _all_legend_labels += tuple(_ls)

        # suppress multiple entries for the same label
        _hs, _ls = [], []
        _seen_labels = set()
        for _h, _l in zip(_all_legend_handles, _all_legend_labels):
            if _l not in _seen_labels:
                _hs.append(_h)
                _ls.append(_l)
                _seen_labels.add(_l)

        _plot_result_dict['legend_handles'] = _hs
        _plot_result_dict['legend_labels'] = _ls

        return _plot_result_dict

    def plot_scatter(self, axes_array):
        """
        Plot a one-dimensional ensemble variable as a matrix of scatter plots.

        :param axes_array: an array of ``matplotlib`` axes objects in which to perform the plotting. The array must
                           be a two-dimensional square (*N* x *N*) array, with *N* being identical to the
                           size of the ensemble variable.

        :type axes_array: ``numpy.ndarray`` of ``matplotlib`` ``Axes`` objects.

        :return: mapping containing plot metadata and other information
        :rtype: `dict`
        """
        axes_array = np.asarray(axes_array)

        if self._var.ndim != 1:
            raise EnsembleError("Cannot create scatter plots: ensemble variable "
                                "must be a 1D array, but got shape {} instead!".format(self._var.shape))

        _ncols = self._var.shape[0]

        if axes_array.shape != (_ncols, _ncols):
            raise EnsembleError("Cannot create scatter plots: "
                                "`axes_array` must be a 2D ({0},{0}) array, "
                                "but got shape {1} instead!".format(self._var.shape[0], axes_array.shape))

        # get the expected mean and its error from the ensemble variable distribution
        _dist = self._var.dist
        if _dist is not None:
            _expected_means = np.atleast_2d(self._var.dist.mean)
            # TODO: presumably only valid/relevant for Gaussian -> solution for non-Gaussian?
            _expected_mean_errors = np.atleast_2d(self._var.dist.std) / np.sqrt(self._var.size)

        _all_legend_handles = []
        _all_legend_labels = []

        _plot_result_dict = dict()
        for _index1, _axes in enumerate(np.atleast_2d(axes_array)):
            for _index2, _ax in enumerate(_axes):
                if _ax is None:
                    continue

                _data_x = self._var.values[:, _index1]
                _data_y = self._var.values[:, _index2]
                assert len(_data_x) == self._var.size
                assert len(_data_y) == self._var.size

                _ = _ax.scatter(
                    _data_x,
                    _data_y,
                    # range=self._value_ranges[_index], # TODO
                    label=self._ensemble_label,
                    marker='.'
                )

                if _dist is not None:
                    _ax.axvline(x=_expected_means[:, _index1],
                                #extra label for expected 'x' mean covered by 'y' mean legend entry
                                **self.__class__._DEFAULT_PLOT_EXPECTED_MEAN_KWARGS)
                    _ax.axhline(y=_expected_means[:, _index2],
                                label="expected mean",
                                **self.__class__._DEFAULT_PLOT_EXPECTED_MEAN_KWARGS)

                if self._value_ranges is not None:
                    _ax.set_xlim(self._value_ranges[0, _index1])
                    _ax.set_ylim(self._value_ranges[0, _index2])

                _ax.xaxis.set_ticks([])
                _ax.yaxis.set_ticks([])

                # set the x axis label
                if self._variable_labels is not None:
                    _xlabel = self._variable_labels[0, _index1]
                    _ylabel = self._variable_labels[0, _index2]
                    if _xlabel is not None:
                        _ax.set_xlabel(_xlabel)
                    if _ylabel is not None:
                        _ax.set_ylabel(_ylabel)

                # collect legend handles and labels
                _hs, _ls = _ax.get_legend_handles_labels()
                _all_legend_handles += tuple(_hs)
                _all_legend_labels += tuple(_ls)

        # suppress multiple entries for the same label
        _hs, _ls = [], []
        _seen_labels = set()
        for _h, _l in zip(_all_legend_handles, _all_legend_labels):
            if _l not in _seen_labels:
                _hs.append(_h)
                _ls.append(_l)
                _seen_labels.add(_l)

        _plot_result_dict['legend_handles'] = _hs
        _plot_result_dict['legend_labels'] = _ls

        return _plot_result_dict



from abc import ABC, abstractmethod
import numpy as np
from scipy.optimize import minimize
import unittest


def fcn_1(x):
    return 42.*(x - 42.)**2 + 42.


def fcn_3(x, y, z):
    return (x - 1.23)**2 + 0.5*(y - 4.32)**2 + 3*(z - 9.81)**2 - 5.23


def fcn_3_wrapper(args):
    return fcn_3(*args)


class AbstractMinimizerTest(ABC):

    @abstractmethod
    def _get_minimizer(self, parameter_names, parameter_values, parameter_errors,
                       function_to_minimize):
        pass

    @property
    @abstractmethod
    def _expected_tolerance(self):
        pass

    # noinspection PyAttributeOutsideInit
    def setUp(self):
        self.func1 = fcn_1
        self.par_names_fcn1 = ['x']
        self.initial_pars_fcn1 = (1.0,)
        self.initial_errs_fcn1 = (0.1,)
        self.m1 = self._get_minimizer(
            parameter_names=self.par_names_fcn1,
            parameter_values=self.initial_pars_fcn1,
            parameter_errors=self.initial_errs_fcn1,
            function_to_minimize=self.func1,
        )

        self.func3 = fcn_3
        self.par_names_fcn3 = ['x', 'y', 'z']
        self.initial_pars_fcn3 = (1.0, 2.0, 3.0)
        self.initial_errs_fcn3 = (0.1, 0.2, 0.5)
        self.m3 = self._get_minimizer(
            parameter_names=self.par_names_fcn3,
            parameter_values=self.initial_pars_fcn3,
            parameter_errors=self.initial_errs_fcn3,
            function_to_minimize=self.func3,
        )

        self._ref_fval_fcn3 = -5.23
        self._ref_par_val_fcn3 = np.array([1.23, 4.32, 9.81])
        self._ref_par_val_fcn3_fix_x = np.array([1.0, 4.32, 9.81])
        self._ref_par_val_fcn3_limit_x = np.array([0.7, 4.32, 9.81])
        self._ref_par_val_fcn3_limit_x_2 = np.array([3.5, 4.32, 9.81])
        self._ref_par_val_fcn3_fix_x_limit_y = np.array([1.0, 0.5, 9.81])
        self._ref_par_err_fcn3 = 1.0 / np.sqrt(np.array([1.0, 0.5, 3.0]))
        self._ref_par_err_fcn3_fix_x = np.array(self._ref_par_err_fcn3)
        self._ref_par_err_fcn3_fix_x[0] = 0
        self._ref_hessian_fcn3 = np.diag(2.0 * np.array([1.0, 0.5, 3.0]))
        self._ref_hessian_inv_fcn3 = np.diag(1.0 / (2.0 * np.array([1.0, 0.5, 3.0])))
        self._ref_cov_mat_fcn3 = np.diag(1.0 / np.array([1.0, 0.5, 3.0]))
        self._ref_cor_mat_fcn3 = np.eye(3)
        self._ref_hessian_fcn3_fix_x = np.array(self._ref_hessian_fcn3)
        self._ref_hessian_fcn3_fix_x[0, 0] = 0
        self._ref_hessian_inv_fcn3_fix_x = np.array(self._ref_hessian_inv_fcn3)
        self._ref_hessian_inv_fcn3_fix_x[0, 0] = 0
        self._ref_cov_mat_fcn3_fix_x = np.array(self._ref_cov_mat_fcn3)
        self._ref_cov_mat_fcn3_fix_x[0, 0] = 0
        self._ref_cor_mat_fcn3_fix_x = np.array(self._ref_cor_mat_fcn3)
        self._ref_cor_mat_fcn3_fix_x[0, 0] = 0

        self._scipy_fmin = minimize(fcn_3_wrapper, np.ones(3), method="bfgs")

        self._ref_contour_m3_x_y_5 = np.array([
            [0.23, 1.23000000, 2.23, 1.23000000, 0.45822021],
            [4.32, 2.90578644, 4.32, 5.73421356, 5.21928411]
        ])
        self._ref_profile_m3_x_5 = np.array([
            [-0.77, 0.23, 1.23, 2.23, 3.23],
            [ 4.00, 1.00, 0.00, 1.00, 4.00]
        ])
        self._ref_profile_m3_x_5_no_subtract_min = np.array(self._ref_profile_m3_x_5)
        self._ref_profile_m3_x_5_no_subtract_min[1] += self._ref_fval_fcn3

    def test_initial_properties_correct(self):
        for _minimizer, _par_names, _initial_pars, _initial_errs, _func in zip(
                [self.m1, self.m3],
                [self.par_names_fcn1, self.par_names_fcn3],
                [self.initial_pars_fcn1, self.initial_pars_fcn3],
                [self.initial_errs_fcn1, self.initial_errs_fcn3],
                [fcn_1, fcn_3]):
            _num_pars = len(_initial_pars)
            self.assertFalse(_minimizer.did_fit)
            self.assertEqual(_minimizer.errordef, 1.0)
            self.assertIs(type(_minimizer.errordef), float)
            self.assertIs(_minimizer.function_to_minimize, _func)
            self.assertEqual(_minimizer.function_value, _func(*_initial_pars))
            self.assertTrue(type(_minimizer.function_value) in [float, np.float64])
            self.assertEqual(_minimizer.num_pars, _num_pars)
            self.assertIs(type(_minimizer.num_pars), int)
            self.assertTrue(np.all(_minimizer.parameter_values == _initial_pars))
            self.assertIs(type(_minimizer.parameter_values), np.ndarray)
            self.assertTrue(np.all(_minimizer.parameter_errors == _initial_errs))
            self.assertIs(type(_minimizer.parameter_errors), np.ndarray)
            self.assertIs(_minimizer.asymmetric_parameter_errors_if_calculated, None)
            self.assertIs(_minimizer.asymmetric_parameter_errors, None)
            self.assertIs(_minimizer.asymmetric_parameter_errors_if_calculated, None)
            self.assertEqual(_minimizer.parameter_names, _par_names)
            self.assertIs(type(_minimizer.parameter_names), list)
            self.assertEqual(_minimizer.tolerance, self._expected_tolerance)
            self.assertIs(type(_minimizer.tolerance), float)
            self.assertIs(_minimizer.hessian, None)
            self.assertIs(_minimizer.hessian_inv, None)
            self.assertIs(_minimizer.cov_mat, None)
            self.assertIs(_minimizer.cor_mat, None)

    def test_properties_copied(self):
        for _minimizer in [self.m1, self.m3]:
            _minimizer.minimize()
            self.assertIsNot(_minimizer.parameter_values, _minimizer.parameter_values)
            self.assertIsNot(_minimizer.parameter_errors, _minimizer.parameter_errors)
            self.assertIsNot(
                _minimizer.asymmetric_parameter_errors, _minimizer.asymmetric_parameter_errors)
            self.assertIsNot(_minimizer.parameter_names, _minimizer.parameter_names)
            self.assertIsNot(_minimizer.hessian, _minimizer.hessian)
            self.assertIsNot(_minimizer.hessian_inv, _minimizer.hessian_inv)
            self.assertIsNot(_minimizer.cov_mat, _minimizer.cov_mat)
            self.assertIsNot(_minimizer.cor_mat, _minimizer.cor_mat)

    def test_neg_par_errs_raise(self):
        for _minimizer in [self.m1, self.m3]:
            with self.assertRaises(ValueError):
                _minimizer.parameter_errors = -_minimizer.parameter_errors

    def test_did_fit_errordef(self):
        for _minimizer in [self.m1, self.m3]:
            self.assertFalse(_minimizer.did_fit)
            _minimizer.minimize()
            self.assertTrue(_minimizer.did_fit)
            _minimizer.errordef = _minimizer.errordef
            self.assertFalse(_minimizer.did_fit)

    def test_did_fit_parameter_values(self):
        for _minimizer in [self.m1, self.m3]:
            self.assertFalse(_minimizer.did_fit)
            _minimizer.minimize()
            self.assertTrue(_minimizer.did_fit)
            _minimizer.parameter_values = _minimizer.parameter_values
            self.assertFalse(_minimizer.did_fit)

    def test_did_fit_parameter_errors(self):
        for _minimizer in [self.m1, self.m3]:
            self.assertFalse(_minimizer.did_fit)
            _minimizer.minimize()
            self.assertTrue(_minimizer.did_fit)
            _minimizer.parameter_errors = _minimizer.parameter_errors
            self.assertFalse(_minimizer.did_fit)

    def test_did_fit_set(self):
        for _minimizer in [self.m1, self.m3]:
            self.assertFalse(_minimizer.did_fit)
            _minimizer.minimize()
            self.assertTrue(_minimizer.did_fit)
            _minimizer.set(_minimizer.parameter_names[0], _minimizer.parameter_values[0])
            self.assertFalse(_minimizer.did_fit)

    def test_did_fit_tolerance(self):
        for _minimizer in [self.m1, self.m3]:
            self.assertFalse(_minimizer.did_fit)
            _minimizer.minimize()
            self.assertTrue(_minimizer.did_fit)
            _minimizer.tolerance = _minimizer.tolerance
            self.assertFalse(_minimizer.did_fit)

    def test_compare_par_values_minimize_fcn1(self):
        self.m1.minimize()
        _par_vals = self.m1.parameter_values
        self.assertIs(type(_par_vals), np.ndarray)
        self.assertAlmostEqual(_par_vals[0], 42., places=6)

    def test_compare_par_errors_minimize_fcn1(self):
        self.m1.minimize()
        _par_errs = self.m1.parameter_errors
        self.assertIs(type(_par_errs), np.ndarray)
        self.assertAlmostEqual(_par_errs[0], np.sqrt(1.0 / 42.0))

    def test_compare_func_value_minimize_fcn1(self):
        self.m1.minimize()
        _fval = self.m1.function_value
        self.assertTrue(type(_fval) in [float, np.float64])
        self.assertAlmostEqual(_fval, 42.0)

    def test_compare_par_values_minimize_fcn3(self):
        self.m3.minimize()
        _par_vals = self.m3.parameter_values
        self.assertIs(type(_par_vals), np.ndarray)
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, _par_vals, rtol=0, atol=1e-6))

    def test_compare_par_errors_minimize_fcn3(self):
        self.m3.minimize()
        _par_errs = self.m3.parameter_errors
        self.assertIs(type(_par_errs), np.ndarray)
        self.assertTrue(np.allclose(
            self._ref_par_err_fcn3, _par_errs, rtol=0, atol=1e-6))

    def test_compare_func_value_minimize_fcn3(self):
        self.m3.minimize()
        _fval = self.m3.function_value
        self.assertTrue(type(_fval) in [float, np.float64])
        self.assertAlmostEqual(_fval, -5.23)

    def test_compare_par_values_minimize_fcn3_fix_release_x(self):
        self.m3.fix('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_fix_x, self.m3.parameter_values, rtol=0, atol=1e-6))

        self.m3.release('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_compare_par_errors_minimize_fcn3_fix_release_x(self):
        self.m3.fix('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_err_fcn3_fix_x, self.m3.parameter_errors, rtol=0, atol=1e-6))

        self.m3.release('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_err_fcn3, self.m3.parameter_errors, rtol=0, atol=1e-6))

    def test_compare_par_values_minimize_fcn3_limit_unlimit_x(self):
        self.m3.limit('x', (-1., 0.7))
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_limit_x, self.m3.parameter_values, rtol=0, atol=1e-6))

        self.m3.unlimit('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_compare_par_values_minimize_fcn3_limit_unlimit_onesided(self):
        self.m3.limit('x', (None, 0.7))
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_limit_x, self.m3.parameter_values, rtol=0, atol=1e-6))

        self.m3.unlimit('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_compare_par_values_minimize_fcn3_limit_unlimit_onesided_2(self):
        self.m3.limit('x', (3.5, None))
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_limit_x_2, self.m3.parameter_values, rtol=0, atol=1e-6))

        self.m3.unlimit('x')
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_limit_x_after_minimize_fcn3(self):
        self.m3.minimize()
        self.m3.limit("x", (-1.0, 0.7))
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_limit_x, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_limit_x_after_minimize_fcn3_2(self):
        self.m3.minimize()
        self.m3.limit("x", (3.5, 10.0))
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_limit_x_2, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_compare_par_values_minimize_fcn3_fix_x_limit_y(self):
        self.m3.fix("x")
        self.m3.limit("y", (-10, 0.5))
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3_fix_x_limit_y, self.m3.parameter_values, rtol=0, atol=1e-6))

        self.m3.release("x")
        self.m3.unlimit("y")
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self._ref_par_val_fcn3, self.m3.parameter_values, rtol=0, atol=1e-6))

    def test_compare_cov_mat_minimize_fcn3(self):
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.cov_mat, self._ref_cov_mat_fcn3, rtol=0, atol=1e-6)
        )
        self.assertAlmostEqual(self.m3.function_value, self._ref_fval_fcn3)
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_cor_mat_minimize_fcn3(self):
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.cor_mat, self._ref_cor_mat_fcn3, rtol=0, atol=1e-6)
        )
        self.assertAlmostEqual(self.m3.function_value, self._ref_fval_fcn3)
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_cov_mat_minimize_fcn3_fix_x(self):
        self.m3.fix('x')
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.cov_mat, self._ref_cov_mat_fcn3_fix_x, rtol=0, atol=1e-6)
        )

    def test_compare_cor_mat_minimize_fcn3_fix_x(self):
        self.m3.fix('x')
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.cor_mat, self._ref_cor_mat_fcn3_fix_x, rtol=0, atol=1e-6)
        )

    def test_compare_cov_mat_minimize_fcn3_fix_x_reverse(self):
        self.m3.minimize()
        self.m3.fix('x')
        self.assertTrue(
            np.allclose(self.m3.cov_mat, self._ref_cov_mat_fcn3_fix_x, rtol=0, atol=2.5e-6)
        )

    def test_compare_cor_mat_minimize_fcn3_fix_x_reverse(self):
        self.m3.minimize()
        self.m3.fix('x')
        self.assertTrue(
            np.allclose(self.m3.cor_mat, self._ref_cor_mat_fcn3_fix_x, rtol=0, atol=1e-6)
        )

    def test_raise_fix_xyz(self):
        self.m3.fix_several(["x", "y", "z"])
        with self.assertRaises(RuntimeError):
            self.m3.minimize()

    def test_compare_par_values_minimize_fcn1_errdef_4(self):
        self.m1.errordef = 4.0
        self.m1.minimize()
        self.assertAlmostEqual(self.m1.parameter_values[0], 42., places=6)

    def test_compare_par_errors_minimize_fcn1_errdef_4(self):
        self.m1.errordef = 4.0
        self.m1.minimize()
        self.assertAlmostEqual(self.m1.parameter_errors[0], np.sqrt(4.0 * (1. / 42.)))

    def test_compare_cov_mat_minimize_fcn3_errdef_4(self):
        self.m3.errordef = 4.0
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.cov_mat, 4.0 * self._ref_cov_mat_fcn3, rtol=0, atol=1e-6)
        )

    def test_compare_hessian_minimize_fcn3(self):
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian, self._ref_hessian_fcn3, rtol=0, atol=1e-6)
        )
        self.assertAlmostEqual(self.m3.function_value, self._ref_fval_fcn3)
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_hessian_minimize_fcn3_fix_x(self):
        self.m3.fix("x")
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian, self._ref_hessian_fcn3_fix_x, rtol=0, atol=1e-6)
        )

    def test_compare_hessian_minimize_fcn3_fix_x_reverse(self):
        self.m3.minimize()
        self.m3.fix("x")
        self.assertTrue(
            np.allclose(self.m3.hessian, self._ref_hessian_fcn3_fix_x, rtol=0, atol=2.5e-6)
        )

    def test_compare_hessian_minimize_fcn3_errdef_4(self):
        self.m3.errordef = 4.0
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian, self._ref_hessian_fcn3, rtol=0, atol=1e-6)
        )

    def test_compare_hessian_inv_minimize_fcn3(self):
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian_inv, self._ref_hessian_inv_fcn3, rtol=0, atol=1e-6)
        )
        self.assertAlmostEqual(self.m3.function_value, self._ref_fval_fcn3)
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_hessian_inv_minimize_fcn3_fix_x(self):
        self.m3.fix("x")
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian_inv, self._ref_hessian_inv_fcn3_fix_x, rtol=0, atol=1e-6)
        )

    def test_compare_hessian_inv_minimize_fcn3_fix_x_reverse(self):
        self.m3.minimize()
        self.m3.fix("x")
        self.assertTrue(
            np.allclose(self.m3.hessian_inv, self._ref_hessian_inv_fcn3_fix_x, rtol=0, atol=2.5e-6)
        )

    def test_compare_hessian_inv_minimize_fcn3_errdef_4(self):
        self.m3.errordef = 4.0
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.hessian_inv, self._ref_hessian_inv_fcn3, rtol=0, atol=1e-6)
        )

    def test_compare_asymm_errs_minimize_fnc1(self):
        self.m1.minimize()
        self.assertAlmostEqual(
            self.m1.asymmetric_parameter_errors[0, 0], -np.sqrt(1.0 / 42.0), places=6)
        self.assertAlmostEqual(
            self.m1.asymmetric_parameter_errors[0, 1], np.sqrt(1.0 / 42.0), places=6)
        self.assertAlmostEqual(self.m1.function_value, 42)
        self.assertAlmostEqual(self.m1.parameter_values[0], 42.)

    def test_compare_asymm_errs_minimize_fnc3(self):
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 0], -self._ref_par_err_fcn3))
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 1], self._ref_par_err_fcn3))
        self.assertAlmostEqual(self.m3.function_value, self._ref_fval_fcn3)
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_asymm_errs_minimize_fnc3_fix_x(self):
        self.m3.fix("x")
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 0], -self._ref_par_err_fcn3_fix_x))
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 1], self._ref_par_err_fcn3_fix_x))
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3_fix_x))

    def test_compare_asymm_errs_minimize_fnc3_fix_x_reverse(self):
        self.m3.minimize()
        self.m3.fix("x")
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 0], -self._ref_par_err_fcn3_fix_x))
        self.assertTrue(np.allclose(
            self.m3.asymmetric_parameter_errors[:, 1], self._ref_par_err_fcn3_fix_x))
        self.assertTrue(np.allclose(self.m3.parameter_values, self._ref_par_val_fcn3))

    def test_compare_par_values_to_scipy_optimize(self):
        self.m3.minimize()
        self.assertTrue(
            np.allclose(self.m3.parameter_values, self._scipy_fmin.x, rtol=0, atol=1e-6)
        )

    @unittest.skip("SLSQP does not compute hessian by default")
    def test_compare_hessian_inv_to_scipy_optimize(self):
        self.m3.minimize()
        _hm_inv = self.m3.hessian_inv
        _hm_inv_scipy = self._scipy_fmin.hess_inv
        self.assertTrue(
            np.allclose(_hm_inv, _hm_inv_scipy, atol=1e-2)
        )

    def test_compare_func_value_set_fcn3_parameters(self):
        self.m3.set('x', 1.23)
        self.m3.set('y', 4.32)
        self.m3.set('z', 9.81)
        self.assertAlmostEqual(self.m3.function_value, -5.23)

    def test_set_raise(self):
        for _minimizer in [self.m1, self.m3]:
            with self.assertRaises(ValueError):
                _minimizer.set("DEADBEEF", 1)

    def test_profile_m3_x(self):
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self.m3.profile('x', bins=5, subtract_min=True),
            self._ref_profile_m3_x_5, atol=1e-7
        ))

    def test_profile_m3_x_no_subtract_min(self):
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self.m3.profile('x', bins=5, subtract_min=False),
            self._ref_profile_m3_x_5_no_subtract_min, atol=1e-7
        ))

    def test_profile_m3_x_fix_y(self):
        self.m3.fix("y")
        self.m3.minimize()
        self.assertTrue(np.allclose(
            self.m3.profile('x', bins=5, subtract_min=True),
            self._ref_profile_m3_x_5, atol=1e-7
        ))

    def test_profile_raise_no_fit(self):
        with self.assertRaises(RuntimeError):
            self.m3.profile("x")

    @unittest.skip("Testing contours not yet implemented!")
    def test_contour_m3_x_y(self):
        self.m3.minimize()
        _cont = self.m3.contour('x', 'y', numpoints=5, sigma=1.0)
        self.assertTrue(
            np.allclose(_cont, self._ref_contour_m3_x_y_5, atol=1e-3)
        )

    def test_contour_raise_no_fit(self):
        with self.assertRaises(RuntimeError):
            self.m3.contour("x", "y")



import abc
import six

from .._base import GenericDReprBase
from ....core.constraint import GaussianSimpleParameterConstraint, GaussianMatrixParameterConstraint

__all__ = ["ConstraintDReprBase"]


@six.add_metaclass(abc.ABCMeta)
class ConstraintDReprBase(GenericDReprBase):
    BASE_OBJECT_TYPE_NAME = 'constraint'

    _CLASS_TO_OBJECT_TYPE_NAME = {
        GaussianSimpleParameterConstraint: 'simple',
        GaussianMatrixParameterConstraint: 'matrix'
    }
    _OBJECT_TYPE_NAME_TO_CLASS = {
        'simple': GaussianSimpleParameterConstraint,
        'matrix': GaussianMatrixParameterConstraint
    }

    def __init__(self, constraint=None):
        self._kafe_object = constraint
        super(ConstraintDReprBase, self).__init__()



"""
Classes for handling of uncertainties in kafe2 fits.
"""

import abc
import numpy as np
import six
import warnings

import logging

logger = logging.getLogger(__name__)
logging.basicConfig()

__all__ = ['CovMat', 'SimpleGaussianError', 'MatrixGaussianError']


def cov_mat_from_float(value, size, correlation=0.0):
    _val = float(value)
    return cov_mat_from_float_list([_val] * size, correlation)


def cov_mat_from_float_list(value_list, correlation=0.0):
    _vals = np.array(list(map(float, value_list)))
    correlation = float(correlation)
    if not (0.0 <= correlation <= 1.0):
        raise ValueError("Correlation must be between 0 and 1: %g given." % (correlation,))
    if correlation == 1.0:
        _mat = np.outer(_vals, _vals)
    elif correlation == 0.0:
        _mat = np.diag(_vals**2)
    else:
        _mat = np.diag(_vals**2 * (1.0 - correlation))
        _mat += np.outer(_vals, _vals) * correlation
    return CovMat(_mat)


# Data structure for Covariance Matrices
class CovMat(object):
    def __init__(self, matrix):
        self.mat = matrix

    # -- 'magic' methods

    def __iadd__(self, other):
        self.mat += other.mat
        return self

    def __add__(self, other):
        _new = CovMat(self.mat)
        _new += other
        return _new

    def __eq__(self, other):
        return np.all(self._mat == other.mat) if isinstance(other, CovMat) \
            else np.all(self._mat == other)

    def __len__(self):
        return self._size

    # -- private methods

    def _invalidate_cache(self):
        self._chol = None
        self._inverse = None
        self._cor_mat = None
        self._cond = None
        self._inverse = None

    # -- public interface

    def rescale(self, old_reference_values, new_reference_values):
        """
        Rescale the covariance matrix to new reference values.
        """
        _old_outer = np.asarray(np.outer(old_reference_values, old_reference_values))
        _new_outer = np.asarray(np.outer(new_reference_values, new_reference_values))
        self._mat *= _new_outer / _old_outer
        self._invalidate_cache()

    @property
    def mat(self):
        """
        Get the covariance matrix.
        """
        return np.array(self._mat)

    @mat.setter
    def mat(self, matrix):
        """
        Set the covariance matrix.
        """
        if isinstance(matrix, CovMat):
            matrix = matrix.mat

        self._mat = np.array(matrix)
        if self._mat.ndim != 2 or self._mat.shape[0] != self._mat.shape[1]:
            raise ValueError(
                "Covariance matrix must be square matrix, shape %r given." % (self._mat.shape,))
        self._size = self._mat.shape[0]
        self._cond = None

        self._invalidate_cache()

    @property
    def cor_mat(self):
        """
        Correlation matrix corresponding to the covariance matrix.
        """
        if self._cor_mat is None:
            _sqrt_vars = np.sqrt(np.diag(self._mat))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._cor_mat = self._mat / np.outer(_sqrt_vars, _sqrt_vars)
        return self._cor_mat

    @property
    def I(self):
        """
        Inverse of the covariance matrix. Returns ``None`` if matrix is singular.
        """
        if self._inverse is None:
            try:
                self._inverse = np.linalg.inv(self._mat)
            except np.linalg.LinAlgError:
                pass  # fail silently if matrix is singular
        return self._inverse

    @property
    def chol(self):
        """
        Lower diagonal matrix resulting from the Cholesky decomposition of the covariance matrix.
        Returns ``None`` if matrix is not positive definite.
        """
        if self._chol is None:
            try:
                self._chol = np.linalg.cholesky(self._mat)
            except np.linalg.LinAlgError:
                pass  # fail silently if matrix is not positive definite
        return self._chol

    @property
    def cond(self):
        """
        Condition number of the matrix.
        """
        if self._cond is None:
            self._cond = np.linalg.cond(self._mat)
        return self._cond

    @property
    def split_svd(self):
        if self.chol is None:
            return None
        _l = []
        _u, _v, _w = np.linalg.svd(self.chol)
        for _sv, _sc in zip(_v, _u.T):
            _l.append(np.outer(_sc, _sc) * _sv**2)
        return _l

    @property
    def split_diag_svd(self):
        _m0 = np.diag(np.diag(self._mat))
        _m = CovMat(self._mat - _m0)
        if _m is None:
            return None
        _l = [_m0]
        _u, _v, _w = np.linalg.svd(self.chol)
        for _sv, _sc in zip(_v, _u.T):
            _l.append(np.outer(_sc, _sc) * _sv ** 2)
        return _l


# Data structures for Gaussian Errors
@six.add_metaclass(abc.ABCMeta)
class GaussianErrorBase(object):
    """
    Purely abstract class. Defines the minimal interface required by all specializations.
    """
    @property
    @abc.abstractmethod
    def error(self):
        """Pointwise error array."""

    @property
    @abc.abstractmethod
    def error_rel(self):
        """Pointwise error array (relative errors)."""

    @property
    def reference(self):
        """Array of reference values for the error."""
        if callable(self._reference):
            return self._reference()
        else:
            return self._reference

    @reference.setter
    def reference(self, reference):
        if reference is None:
            ##self._reference = np.ones_like(self._err_val)
            self._reference = None
        elif callable(reference):
            self._reference = reference
        else:
            _ref = np.asarray(reference, dtype=float)
            # check for zero-valued references if error is marked 'relative'
            if self.relative and np.any(_ref == 0):
                warnings.warn(
                    "Relative error has a reference that contains values equal to zero: %s"
                    % _ref
                )
            self._reference = _ref

        # invalidate error_structures opposite declared relativity type
        if self.relative:
            self._cov_mat = None
            self._err = None
        else:
            self._cov_mat_rel = None
            self._err_rel = None

    @property
    @abc.abstractmethod
    def cov_mat(self):
        """Full absolute covariance matrix for error."""

    @property
    @abc.abstractmethod
    def cov_mat_rel(self):
        """Full relative covariance matrix for error."""

    @property
    @abc.abstractmethod
    def cor_mat(self):
        """Correlation matrix for error."""

    # TODO: remove _uncor/_cor from base interface?

    @property
    @abc.abstractmethod
    def error_uncor(self):
        """Pointwise array of 'uncorrelated' parts of absolute errors."""

    @property
    @abc.abstractmethod
    def error_cor(self):
        """Pointwise array of 'correlated' parts of absolute errors."""

    @property
    @abc.abstractmethod
    def error_rel_uncor(self):
        """Pointwise array of 'uncorrelated' parts of relative errors."""

    @property
    @abc.abstractmethod
    def error_rel_cor(self):
        """Pointwise array of 'correlated' parts of relative errors."""

    @property
    @abc.abstractmethod
    def cov_mat_uncor(self):
        """'Uncorrelated' part of absolute covariance matrix for error."""

    @property
    @abc.abstractmethod
    def cov_mat_cor(self):
        """'Fully correlated' part of absolute covariance matrix for error."""

    @property
    @abc.abstractmethod
    def cov_mat_rel_uncor(self):
        """'Uncorrelated' part of relative covariance matrix for error."""

    @property
    @abc.abstractmethod
    def cov_mat_rel_cor(self):
        """'Fully correlated' part of relative covariance matrix for error."""

    @property
    @abc.abstractmethod
    def fit_indices(self):
        """Indices of fits that have this error when used inside a MultiFit."""

    def get_cov_mat_object(self):
        """
        Returns the internally used `CovMat` object used to represent measurement errors. (advanced)
        """
        return self._cov_mat


class SimpleGaussianError(GaussianErrorBase):
    """
    A Gaussian Error constructed from an array of uncertainties, one for each data point, and a single
    non-negative correlation coefficient indicating the correlation of any two data points.
    This object constructs a covariance matrix object (py:obj:`CovMat`) from this information.

    An error object may be constructed as 'absolute' (default) or 'relative'. In the latter case, the
    uncertainty values are assumed to be given relative to the reference values (measurement or theory).
    There can be optionally specified using the 'reference' property.

    If an error object is declared as 'absolute' ('relative') and no 'reference' is set, then only 'absolute'
    ('relative') error arrays and covariance matrices are available. If 'reference' is set, these values are
    used to convert 'absolute' ('relative') error arrays or covariance matrices to 'relative' ('absolute') ones.

    """
    def __init__(self, err_val, corr_coeff, relative=False, reference=None, fit_indices=None):
        if not (0.0 <= corr_coeff <= 1.0):
            raise ValueError("Correlation must be between 0 and 1, %g given," % (corr_coeff,))
        err_val = np.asarray(err_val)
        if err_val.ndim != 1:
            raise ValueError(
                f"Error must be one-dimensional but received array with {err_val.ndim} dimensions.")
        self._corr_coeff = float(corr_coeff)
        self._is_relative = relative
        self.reference = reference
        self._fit_indices = fit_indices
        if self.relative:
            self.error_rel = err_val
        else:
            self.error = err_val

    # -- static methods

    @staticmethod
    def _calculate_cov_mat_generic(error_array, corr_coeff):
        """Calculate a covariance matrix from an array of error values and a global correlation coefficient."""
        if corr_coeff > 0:
            cov_mat_uncor_part = np.diag(error_array ** 2 * (1.0 - corr_coeff))
            cov_mat_cor_part = np.outer(error_array, error_array) * corr_coeff
        else:
            cov_mat_uncor_part = np.diag(error_array ** 2)
            cov_mat_cor_part = np.zeros_like(cov_mat_uncor_part)

        cov_mat = CovMat(cov_mat_uncor_part + cov_mat_cor_part)

        return cov_mat, cov_mat_uncor_part, cov_mat_cor_part

    # -- private methods

    def _calculate_cov_mat(self):
        """Calculate absolute covariance matrix for error object."""
        if self.relative:
            if self.reference is None:
                raise AttributeError("Requested 'absolute' errors for error object declared 'relative', but 'reference' not set!")
            _abs_err = self.error_rel * self.reference
        else:
            _abs_err = self.error

        self._cov_mat, self._cov_mat_uncor_part, self._cov_mat_cor_part = self._calculate_cov_mat_generic(_abs_err, self._corr_coeff)

    def _calculate_cov_mat_rel(self):
        """Calculate relative covariance matrix for error object."""
        _rel_err = self.error_rel
        self._cov_mat_rel, self._cov_mat_rel_uncor_part, self._cov_mat_rel_cor_part \
            = self._calculate_cov_mat_generic(_rel_err, self._corr_coeff)

    # -- public methods

    @property
    def relative(self):
        """Returns ``True`` if error is marked as a relative error."""
        return self._is_relative

    # @relative.setter
    # def relative(self, relative):
    #     """Set to ``True`` to mark as a relative error."""
    #     self._is_relative = relative

    @property
    def error(self):
        # calculate relative error from absolute error (if 'reference' is set)
        if self.relative:
            if self.reference is None:
                raise AttributeError(
                    "Requested 'absolute' errors for error object declared 'relative', but 'reference' not set!")
            self._err = self._err_rel * np.abs(self.reference)
        return self._err

    @error.setter
    def error(self, err_val):
        err_val = np.array(err_val, dtype=float)
        if np.any(err_val < 0):
            raise ValueError("Error values must be >= 0. Received: %s" % err_val)
        if self.relative:
            if self.reference is None:
                raise AttributeError(
                    "Setting 'absolute' errors for error object declared 'relative', but 'reference' not set!")

            self._err = err_val
            self._err_rel = err_val / np.abs(self.reference)
        else:
            self._err = err_val
            self._err_rel = None

        # invalidate cov mats
        self._cov_mat = None
        self._cov_mat_rel = None

    @property
    def error_uncor(self):
        # TODO: cache
        return self.error * np.sqrt(1.0 - self._corr_coeff)

    @property
    def error_cor(self):
        # TODO: cache
        return self.error * np.sqrt(self._corr_coeff)

    @property
    def error_rel_cor(self):
        # TODO: cache
        return self.error_rel * np.sqrt(self._corr_coeff)

    @property
    def error_rel_uncor(self):
        # TODO: cache
        return self.error_rel * np.sqrt(1.0 - self._corr_coeff)

    @property
    def error_rel(self):
        # calculate relative error from absolute error (if 'reference' is set)
        if not self.relative:
            if self.reference is None:
                raise AttributeError(
                    "Requested 'relative' errors for error object declared 'absolute', but 'reference' not set!")
            self._err_rel = self._err / np.abs(self.reference)
        return self._err_rel

    @error_rel.setter
    def error_rel(self, err_val):
        err_val = np.array(err_val, dtype=float)
        if np.any(err_val < 0):
            raise ValueError("Error values must be >= 0. Received: %s" % err_val)
        if self.relative:
            self._err = None
            self._err_rel = err_val
        else:
            if self.reference is None:
                raise AttributeError(
                    "Setting 'absolute' errors for error object declared 'relative', but 'reference' not set!")
            self._err = err_val * self.reference
            self._err_rel = err_val

        # invalidate cov mats
        self._cov_mat = None
        self._cov_mat_rel = None

    @property
    def cov_mat(self):
        if self._cov_mat is None:
            self._calculate_cov_mat()
        return self._cov_mat.mat

    @property
    def cov_mat_inverse(self):
        if self._cov_mat is None:
            self._calculate_cov_mat()
        return self._cov_mat.I

    @property
    def cov_mat_uncor(self):
        if self._cov_mat is None:
            self._calculate_cov_mat()
        return self._cov_mat_uncor_part

    @property
    def cov_mat_cor(self):
        if self._cov_mat is None:
            self._calculate_cov_mat()
        return self._cov_mat_cor_part

    @property
    def cov_mat_rel(self):
        if self._cov_mat_rel is None:
            self._calculate_cov_mat_rel()
        return self._cov_mat_rel.mat

    @property
    def cov_mat_rel_inverse(self):
        if self._cov_mat is None:
            self._calculate_cov_mat_rel()
        return self._cov_mat_rel.I

    @property
    def cov_mat_rel_uncor(self):
        if self._cov_mat_rel is None:
            self._calculate_cov_mat_rel()
        return self._cov_mat_rel_uncor_part

    @property
    def cov_mat_rel_cor(self):
        if self._cov_mat_rel is None:
            self._calculate_cov_mat_rel()
        return self._cov_mat_rel_cor_part

    @property
    def cor_mat(self):
        if self.relative:
            _ = self.cov_mat_rel  # call to calculate CovMat
            return self._cov_mat_rel.cor_mat
        else:
            _ = self.cov_mat  # call to calculate CovMat
            return self._cov_mat.cor_mat

    @property
    def corr_coeff(self):
        return self._corr_coeff

    @property
    def fit_indices(self):
        return self._fit_indices


class MatrixGaussianError(GaussianErrorBase):
    """
    A Gaussian Error constructed from a covariance matrix, or a correlation matrix together with
    a pointwise error array.

    An error object may be constructed as 'absolute' (default) or 'relative'. In the latter case, the
    uncertainty values are assumed to be given relative to the reference values (measurement or theory).
    There can be optionally specified using the 'reference' property.

    If an error object is declared as 'absolute' ('relative') and no 'reference' is set, then only 'absolute'
    ('relative') error arrays and covariance matrices are available. If 'reference' is set, these values are
    used to convert 'absolute' ('relative') error arrays or covariance matrices to 'relative' ('absolute') ones.

    """
    def __init__(self, err_matrix, matrix_type, err_val=None, relative=False,
                 reference=None, fit_indices=None):
        err_matrix = np.asarray(err_matrix)
        err_val = err_val if err_val is None else np.asarray(err_val)
        if err_matrix.ndim != 2:
            raise ValueError(
                "Error matrix must be two-dimensional but received "
                f"array with {err_matrix.ndim} dimensions.")
        if err_val is not None and err_val.ndim > 1:
            raise ValueError(
                "Error array must be scalar or one-dimensional but received "
                f"array with {err_val.ndim} dimensions.")
        self._is_relative = relative
        self.reference = reference
        self._fit_indices = fit_indices

        self._err = None
        self._err_rel = None
        self._cov_mat = None
        self._cov_mat_rel = None

        # set the main matrix
        if matrix_type.lower() in ('covariance', 'cov'):
            if err_val is not None:
                raise ValueError("Cannot provide err_val when constructing a covariance matrix!")
            self._matrix_type_at_construction = 'covariance'
            if self.relative:
                self.cov_mat_rel = err_matrix
            else:
                self.cov_mat = err_matrix
        elif matrix_type.lower() in ('correlation', 'correlations', 'cor', 'corr'):
            self._matrix_type_at_construction = 'correlation'
            if err_val is None:
                raise ValueError(
                    "Cannot construct matrix-type error from correlation matrix "
                    "without an array of error values!")
            _cm = self._calculate_cov_mat_from_cor_mat_and_error_array(err_val, err_matrix)
            if self.relative:
                self.cov_mat_rel = _cm
            else:
                self.cov_mat = _cm
        else:
            raise ValueError("Unknown matrix type '%s'. Expected one of: %r" % (matrix_type, ('cov', 'cor')))

    # -- static methods

    @staticmethod
    def _calculate_cov_mat_from_cor_mat_and_error_array(error_array, corr_mat):
        """Calculate a covariance matrix from an array of error values and a correlation matrix."""
        # check if corr_mat has ones on diagonal
        if not np.allclose(np.diag(corr_mat), 1.0):
            raise ValueError("Corelation matrix has non-unit entry on diagonal!")
        # TODO: check if corr_mat is symmetric and positive definite (?)
        if error_array.ndim > 0 and error_array.shape[0] != corr_mat.shape[0]:
            raise ValueError(
                f"Error array has size {error_array.shape[0]} but "
                f"correlation matrix has size {corr_mat.shape[0]}, must be the same.")
        cov_mat = np.asarray(np.outer(error_array, error_array)) * np.asarray(corr_mat)
        return CovMat(cov_mat)

    @staticmethod
    def _calculate_cov_mat_rel_from_cov(cov_mat, reference):
        _ref = np.asarray(reference)
        _refmat = np.outer(_ref, _ref)
        _mat = np.asarray(cov_mat)
        return CovMat(_mat / _refmat)

    @staticmethod
    def _calculate_cov_mat_from_cov_rel(cov_mat_rel, reference):
        _ref = np.asarray(reference)
        _refmat = np.outer(_ref, _ref)
        _mat_rel = np.asarray(cov_mat_rel)
        return CovMat(_mat_rel * _refmat)

    # -- public methods

    @property
    def relative(self):
        """Returns ``True`` if error is marked as a relative error."""
        return self._is_relative

    @property
    def cov_mat(self):
        """"""
        if self.relative:
            if self.reference is None:
                raise AttributeError(
                    "Requested 'absolute' covariance matrix for error object declared 'relative', "
                    "but 'reference' not set!")
            self._cov_mat = self._calculate_cov_mat_from_cov_rel(self.cov_mat_rel, self.reference)
        return self._cov_mat.mat

    @cov_mat.setter
    def cov_mat(self, cov_mat):
        """"""
        self._cov_mat = CovMat(cov_mat)
        self._cov_mat_rel = None

    @property
    def cov_mat_rel(self):
        """"""
        if not self.relative:
            if self.reference is None:
                raise AttributeError(
                    "Requested 'relative' covariance matrix for error object declared 'absolute', but 'reference' not set!")
            self._cov_mat_rel = self._calculate_cov_mat_rel_from_cov(self.cov_mat, self.reference)
        return self._cov_mat_rel.mat

    @cov_mat_rel.setter
    def cov_mat_rel(self, cov_mat_rel):
        """"""
        self._cov_mat_rel = CovMat(cov_mat_rel)
        self._cov_mat = None

    @property
    def cov_mat_inverse(self):
        _ = self.cov_mat  # call to initialize
        return self._cov_mat.I

    @property
    def error(self):
        """"""
        if self._err is None:
            if self.relative:
                if self.reference is None:
                    raise AttributeError(
                        "Requested 'absolute' error array for error object declared 'relative', but 'reference' not set!")
            self._err = np.sqrt(np.diag(self.cov_mat))
        return self._err

    @property
    def error_rel(self):
        """"""
        if self._err_rel is None:
            if not self.relative:
                if self.reference is None:
                    raise AttributeError(
                        "Requested 'relative' error array for error object declared 'absolute', but 'reference' not set!")
            self._err_rel = np.sqrt(np.diag(self.cov_mat_rel))
        return self._err_rel

    @property
    def error_uncor(self):
        raise AttributeError("Cannot get the uncorrelated part of a 'matrix-type' error!")

    @property
    def error_cor(self):
        raise AttributeError("Cannot get the correlated part of a 'matrix-type' error!")

    @property
    def error_rel_cor(self):
        raise AttributeError("Cannot get the correlated part of a 'matrix-type' error!")

    @property
    def error_rel_uncor(self):
        raise AttributeError("Cannot get the uncorrelated part of a 'matrix-type' error!")

    @property
    def cov_mat_uncor(self):
        raise AttributeError("Cannot get the uncorrelated part of a 'matrix-type' error!")

    @property
    def cov_mat_cor(self):
        raise AttributeError("Cannot get the correlated part of a 'matrix-type' error!")

    @property
    def cov_mat_rel_uncor(self):
        raise AttributeError("Cannot get the uncorrelated part of a 'matrix-type' error!")

    @property
    def cov_mat_rel_cor(self):
        raise AttributeError("Cannot get the correlated part of a 'matrix-type' error!")

    @property
    def cor_mat(self):
        _ = self.cov_mat  # call to initialize
        return self._cov_mat.cor_mat

    @property
    def fit_indices(self):
        return self._fit_indices

    # For performance reasons this check is called manually from outside
    def check_cov_mat_symmetry(self):
        _mat = self.cov_mat
        if np.any(_mat != _mat.T):
            warnings.warn("Covariance matrix is not symmetrical: %s" % _mat)




import unittest
import numpy as np

from kafe2.fit import XYContainer, XYParametricModel
from kafe2.fit._base import ModelFunctionBase
from kafe2.core.error import cov_mat_from_float_list



class TestDatastoreXY(unittest.TestCase):

    def setUp(self):
        self._ref_x_data = [0, 1, 2, 3, 4]
        self._ref_y_data = [3.3, 5.5, 2.2, 8.5, 10.2]
        self.data_xy = XYContainer(x_data=self._ref_x_data, y_data=self._ref_y_data)

        self._ref_x_err_abs_singlevalue = 0.1
        self._ref_y_err_abs_singlevalue = 1.2
        self._ref_x_err_abs_valuearray = np.array([0.1, 0.1, 0.2, 0.4, 0.5])
        self._ref_y_err_abs_valuearray = [1.2, 1.1, 1.2, 1.2, 1.1]

        self._ref_x_err_corr_coeff = 0.1
        self._ref_y_err_corr_coeff = 0.23

        self.data_xy.add_error('x', self._ref_x_err_abs_valuearray, name='MyXError', correlation=self._ref_x_err_corr_coeff,
                               relative=False)
        self.data_xy.add_error('y', self._ref_y_err_abs_valuearray, name='MyYError', correlation=self._ref_y_err_corr_coeff,
                               relative=False)

        self._ref_x_cov_mat = cov_mat_from_float_list(self._ref_x_err_abs_valuearray,
                                                      correlation=self._ref_x_err_corr_coeff).mat
        self._ref_y_cov_mat = cov_mat_from_float_list(self._ref_y_err_abs_valuearray,
                                                      correlation=self._ref_y_err_corr_coeff).mat

    def test_get_matching_error_all_empty_dict(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict())
        self.assertEqual(len(_errs), 2)
        self.assertIs(self.data_xy._error_dicts['MyXError']['err'], _errs['MyXError'])
        self.assertIs(self.data_xy._error_dicts['MyYError']['err'], _errs['MyYError'])

    def test_get_matching_error_all_None(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=None)
        self.assertEqual(len(_errs), 2)
        self.assertIs(self.data_xy._error_dicts['MyXError']['err'], _errs['MyXError'])
        self.assertIs(self.data_xy._error_dicts['MyYError']['err'], _errs['MyYError'])

    def test_get_matching_error_name(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(name='MyXError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.data_xy._error_dicts['MyXError']['err'], _errs['MyXError'])

    def test_get_matching_error_type_simple(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(self.data_xy._error_dicts['MyXError']['err'], _errs['MyXError'])
        self.assertIs(self.data_xy._error_dicts['MyYError']['err'], _errs['MyYError'])

    def test_get_matching_error_type_matrix(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_uncorrelated(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(correlated=False))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 2)
        self.assertIs(self.data_xy._error_dicts['MyXError']['err'], _errs['MyXError'])
        self.assertIs(self.data_xy._error_dicts['MyYError']['err'], _errs['MyYError'])

    def test_get_matching_error_axis(self):
        _errs = self.data_xy.get_matching_errors(matching_criteria=dict(axis=1))
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.data_xy._error_dicts['MyYError']['err'], _errs['MyYError'])

    def test_compare_error_reference(self):
        for _err_dict in self.data_xy._error_dicts.values():
            _err_ref_vals = _err_dict['err'].reference
            _axis = _err_dict['axis']
            assert _axis in (0, 1)
            if _axis == 0:
                self.assertTrue(
                    np.all(_err_ref_vals == self._ref_x_data)
                )
            elif _axis == 1:
                self.assertTrue(
                    np.all(_err_ref_vals == self._ref_y_data)
                )


    def test_compare_ref_x_data(self):
        self.assertTrue(np.all(self.data_xy.x == self._ref_x_data))

    def test_compare_ref_y_data(self):
        self.assertTrue(np.all(self.data_xy.y == self._ref_y_data))

    def test_compare_ref_x_err(self):
        self.assertTrue(np.allclose(self.data_xy.x_err, self._ref_x_err_abs_valuearray, atol=1e-10))

    def test_compare_ref_y_err(self):
        self.assertTrue(np.allclose(self.data_xy.y_err, self._ref_y_err_abs_valuearray, atol=1e-10))

    # def test_compare_ref_x_cov_mat(self):
    #     self.assertTrue(np.allclose(self.data_xy.???, self._ref_x_cov_mat, atol=1e-5))
    #
    # def test_compare_ref_y_cov_mat(self):
    #     self.assertTrue(np.allclose(self.data_xy.???, self._ref_y_cov_mat, atol=1e-5))

    def test_compare_ref_total_x_cov_mat(self):
        _err = self.data_xy.get_total_error(0)
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_x_cov_mat, atol=1e-5))

    def test_compare_ref_total_y_cov_mat(self):
        _err = self.data_xy.get_total_error(1)
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_y_cov_mat, atol=1e-5))

    def test_compare_ref_total_y_for_x_plus_y_as_y_err(self):
        self.data_xy.add_error('y', self._ref_x_err_abs_valuearray,
                               correlation=self._ref_x_err_corr_coeff,
                               relative=False)
        _err = self.data_xy.get_total_error(1)
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_y_cov_mat + self._ref_x_cov_mat, atol=1e-5))

    def test_compare_ref_total_y_for_x_plus_y_as_y_err_disabled(self):
        self.data_xy.add_error('y', self._ref_x_err_abs_valuearray,
                               name="MyNewYError",
                               correlation=self._ref_x_err_corr_coeff,
                               relative=False)
        self.data_xy.disable_error("MyNewYError")
        _err = self.data_xy.get_total_error(1)
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_y_cov_mat, atol=1e-5))

    def test_compare_ref_total_y_for_x_plus_y_as_y_err_disabled_reenabled(self):
        self.data_xy.add_error('y', self._ref_x_err_abs_valuearray,
                               name="MyNewYError",
                               correlation=self._ref_x_err_corr_coeff,
                               relative=False)
        self.data_xy.disable_error("MyNewYError")
        _err = self.data_xy.get_total_error(1)
        self.data_xy.enable_error("MyNewYError")
        _err = self.data_xy.get_total_error(1)
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_y_cov_mat + self._ref_x_cov_mat, atol=1e-5))

    def test_raise_add_same_error_name_twice(self):
        self.data_xy.add_error('y', 0.1,
                               name="MyNewYError",
                               correlation=0, relative=False)
        with self.assertRaises(ValueError):
            self.data_xy.add_error('y', 0.1,
                                   name="MyNewYError",
                                   correlation=0, relative=False)

    def test_raise_get_inexistent_error(self):
        with self.assertRaises(ValueError):
            self.data_xy.get_error("MyInexistentYError")

    def test_raise_add_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.data_xy.add_error("x", [1, 2, 3])
        with self.assertRaises(ValueError):
            self.data_xy.add_error("y", [1, 2, 3])

    def test_raise_add_matrix_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("x", [[1, 0], [0, 4]], "cov")
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("y", [[1, 0], [0, 4]], "cov")
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("x", np.eye(5), "cor", err_val=np.ones(10))
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("y", np.eye(5), "cor", err_val=np.ones(10))
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("x", np.eye(5), "cor", err_val=np.eye(5))
        with self.assertRaises(ValueError):
            self.data_xy.add_matrix_error("y", np.eye(5), "cor", err_val=np.eye(5))

    def test_raise_bad_data_shape(self):
        with self.assertRaises(ValueError):
            _ = XYContainer(x_data=[[1, 2], [3, 4]], y_data=[1, 2])
        with self.assertRaises(ValueError):
            _ = XYContainer(x_data=[1, 2], y_data=[[1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            _ = XYContainer(x_data=[1, 2, 3], y_data=[1, 2])


class TestDatastoreXYParametricModel(unittest.TestCase):

    @staticmethod
    def _ref_model_func(x, slope, intercept):
        return slope * x + intercept

    def _ref_model_func_deriv_by_x(x, slope, intercept):
        return slope

    _ref_model_func_deriv_by_x = np.vectorize(_ref_model_func_deriv_by_x)
    _ref_model_func_deriv_by_x = staticmethod(_ref_model_func_deriv_by_x)

    def _ref_model_func_deriv_by_pars(x, slope, intercept):
        return [x, 1]

    _ref_model_func_deriv_by_pars = np.vectorize(_ref_model_func_deriv_by_pars)
    _ref_model_func_deriv_by_pars = staticmethod(_ref_model_func_deriv_by_pars)

    def setUp(self):
        self._ref_x = np.linspace(-5, 5, 11)
        self._test_x = np.linspace(-3, 45, 20)

        self._ref_params = (1.2, 3.3)
        self._ref_data = self._ref_model_func(self._ref_x, *self._ref_params)

        self.xy_param_model = XYParametricModel(
            x_data=self._ref_x, 
            model_func=ModelFunctionBase(self._ref_model_func),
            model_parameters=self._ref_params)

        self._test_params = (3.4, -5.23)
        self._ref_data_ref_x_test_params =  self._ref_model_func(self._ref_x, *self._test_params)
        self._ref_data_test_x_ref_params =  self._ref_model_func(self._test_x, *self._ref_params)
        self._ref_data_test_x_test_params = self._ref_model_func(self._test_x, *self._test_params)



    def test_compare_ref_data(self):
        self.assertTrue(np.all(self.xy_param_model.y == self._ref_data))

    def test_deriv_by_x(self):
        self.assertTrue(
            np.allclose(
                self.xy_param_model.eval_model_function_derivative_by_x(),
                self._ref_model_func_deriv_by_x(self._ref_data, *self._ref_params)
            )
        )

    def test_deriv_by_par(self):
        _dp = [self._ref_model_func_deriv_by_pars(x, *self._ref_params) for x in self._ref_x]
        _dp = np.array(_dp).T
        self.assertTrue(
            np.allclose(
                self.xy_param_model.eval_model_function_derivative_by_parameters(),
                _dp
            )
        )


    def test_change_parameters_test_data(self):
        self.xy_param_model.parameters = self._test_params
        self.assertTrue(np.allclose(self.xy_param_model.y, self._ref_data_ref_x_test_params))

    def test_change_x_test_data(self):
        self.xy_param_model.x = self._test_x
        self.assertTrue(np.allclose(self.xy_param_model.y, self._ref_data_test_x_ref_params))

    def test_change_x_change_parameters_test_data(self):
        self.xy_param_model.x = self._test_x
        self.xy_param_model.parameters = self._test_params
        self.assertTrue(np.allclose(self.xy_param_model.y, self._ref_data_test_x_test_params))

    def test_change_parameters_change_x_test_data(self):
        self.xy_param_model.parameters = self._test_params
        self.xy_param_model.x = self._test_x
        self.assertTrue(np.allclose(self.xy_param_model.y, self._ref_data_test_x_test_params))

    def test_raise_set_data(self):
        with self.assertRaises(TypeError):
            self.xy_param_model.data = self._ref_data_ref_x_test_params

    def test_raise_set_y(self):
        with self.assertRaises(TypeError):
            self.xy_param_model.y = self._ref_data_ref_x_test_params



from __future__ import print_function

import abc
from copy import copy

import numpy as np
import six

from ..io.file import FileIOMixin
from ...core.error import SimpleGaussianError, MatrixGaussianError
from ...tools import random_alphanumeric  # relative import of kafe2.tools not kafe2.fit.tools

__all__ = ["DataContainerBase"]


@six.add_metaclass(abc.ABCMeta)
class DataContainerBase(FileIOMixin):
    """This is a purely abstract class implementing the minimal interface required by all types of data containers.

    It stores measurement data and uncertainties.
    """

    def __init__(self):
        self._error_dicts = dict()
        self._total_error = None
        self._label = None
        self._axis_labels = (None, None)
        self._on_error_change_callback = None
        super(DataContainerBase, self).__init__()

    # -- private methods

    @classmethod
    def _get_base_class(cls):
        return DataContainerBase

    @classmethod
    def _get_object_type_name(cls):
        return 'container'

    @abc.abstractmethod
    def _calculate_total_error(self):
        pass

    @abc.abstractmethod
    def _clear_total_error_cache(self):
        pass

    def _add_error_object(self, name, error_object, **additional_error_dict_keys):
        """create a new entry <name> under self._error_dicts,
        with keys err=<ErrorObject> and arbitrary additional keys"""
        if error_object.error.shape[0] != self.size:
            raise ValueError(f"Error must have size {self.size} but"
                             f"received error with size {error_object.error.shape[0]}")
        _name = name
        if _name is not None and _name in self._error_dicts:
            raise ValueError("Cannot create error source with name '{}': there is already an error "
                             "source registered under that name!".format(_name))
        # be paranoid about name collisions
        while _name is None or _name in self._error_dicts:
            _name = random_alphanumeric(size=8)

        additional_error_dict_keys.setdefault('enabled', True)  # enable error source, unless explicitly disabled

        axis = additional_error_dict_keys.get("axis", None)
        if axis is not None:
            if axis == "x":
                additional_error_dict_keys["axis"] = 0
            elif axis == "y":
                additional_error_dict_keys["axis"] = 1

        _new_err_dict = dict(err=error_object, **additional_error_dict_keys)
        self._error_dicts[_name] = _new_err_dict
        self._on_error_change()
        return _name

    def _get_error_by_name_raise(self, error_name):
        """return a dictionary containing the error object for error 'name' and additional information"""
        _err_dict = self._error_dicts.get(error_name, None)
        if _err_dict is None:
            raise ValueError("No error with name '{}'!".format(error_name))
        return _err_dict

    def _on_error_change(self):
        self._clear_total_error_cache()
        if self._on_error_change_callback is not None:
            self._on_error_change_callback()

    @property
    def label(self):
        """The label describing the dataset.

        :rtype: str or None
        """
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def axis_labels(self):
        """The axis labels describing the dataset.

        :rtype: tuple[str or None, str or None]
        """
        return self._axis_labels

    @axis_labels.setter
    def axis_labels(self, labels):
        self._axis_labels = labels

    @property
    def x_label(self):
        """The x-axis label.

        :rtype: str or None
        """
        return self._axis_labels[0]

    @x_label.setter
    def x_label(self, label):
        _, _y_label = self._axis_labels
        self._axis_labels = (label, _y_label)

    @property
    def y_label(self):
        """The y-axis label.

        :rtype: str or None
        """
        return self._axis_labels[1]

    @y_label.setter
    def y_label(self, label):
        _x_label, _ = self._axis_labels
        self._axis_labels = (_x_label, label)

    @property
    @abc.abstractmethod
    def size(self):
        """The size of the data (number of measurement points).

        :rtype: int
        """
        return 0

    @property
    @abc.abstractmethod
    def data(self):
        """A numpy array containing the data values.

        :rtype: numpy.ndarray[float]
        """
        return np.empty(tuple())

    @property
    @abc.abstractmethod
    def err(self):
        """A numpy array containing the pointwise data uncertainties.

        :rtype: numpy.ndarray[float]
        """
        return np.empty(tuple())

    @property
    @abc.abstractmethod
    def cov_mat(self):
        """A numpy matrix containing the covariance matrix of the data.

        :rtype: numpy.ndarray[numpy.ndarray[float]]"""
        return np.array(np.empty(tuple()))

    @property
    @abc.abstractmethod
    def cov_mat_inverse(self):
        """A numpy matrix containing inverse of the data covariance matrix (or :py:obj`None` if not invertible).

        :rtype: numpy.ndarray[numpy.ndarray[float]] or None
        """
        return np.array(np.empty(tuple()))

    @property
    def has_errors(self):
        """:py:obj:`True` if at least one uncertainty source is defined for the data container.

        :rtype: bool
        """
        return True if self._error_dicts else False

    # -- public methods

    # error-related methods

    def add_error(self, err_val, name=None, correlation=0, relative=False, reference=None):
        """Add an uncertainty source to the data container.

        :param err_val: Pointwise uncertainty/uncertainties for all data points.
        :type err_val: float or numpy.ndarray[float]
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name of the error source will be
                     set to a random alphanumeric string.
        :type name: str or None
        :param correlation: Correlation coefficient between any two distinct data points.
        :type correlation: float
        :param relative: If :py:obj:`True`, **err_val** will be interpreted as a *relative* uncertainty.
        :type relative: bool
        :param reference: The data values to use when computing absolute errors from relative ones (and vice-versa)
        :type reference: typing.Iterable[float] or None
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        try:
            err_val.ndim  # will raise if simple float
        except AttributeError:
            err_val = np.asarray(err_val, dtype=float)

        if err_val.ndim == 0:  # if dimensionless numpy array (i.e. float64), add a dimension
            err_val = np.ones(self.size) * err_val

        _err = SimpleGaussianError(err_val=err_val, corr_coeff=correlation,
                                   relative=relative, reference=reference)

        _name = self._add_error_object(name=name, error_object=_err)
        return _name

    def add_matrix_error(self, err_matrix, matrix_type,
                         name=None, err_val=None, relative=False, reference=None):
        """Add a matrix uncertainty source to the data container.

        :param err_matrix: Covariance or correlation matrix.
        :param matrix_type: One of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``.
        :type matrix_type: str
        :param name: Unique name for this uncertainty source. If :py:obj`None`, the name of the error source will be set to a
                     random alphanumeric string.
        :type name: str or None
        :param err_val: The pointwise uncertainties (mandatory if only a correlation matrix is given).
        :type err_val: typing.Iterable[float]
        :param relative: If :py:obj:`True`, the covariance matrix and/or **err_val** will be interpreted as a *relative*
                         uncertainty.
        :type relative: bool
        :param reference: the data values to use when computing absolute errors from relative ones (and vice-versa)
        :type reference: typing.Iterable[float] or None
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        _err = MatrixGaussianError(err_matrix=err_matrix, matrix_type=matrix_type, err_val=err_val,
                                   relative=relative, reference=reference)
        _err.check_cov_mat_symmetry()

        _name = self._add_error_object(name=name, error_object=_err)
        return _name

    def disable_error(self, error_name):
        """Temporarily disable an uncertainty source so that it doesn't count towards calculating the total uncertainty.

        :param error_name: error name
        :type error_name: str
        """
        _err_dict = self._get_error_by_name_raise(error_name)
        _err_dict['enabled'] = False
        self._on_error_change()

    def enable_error(self, error_name):
        """(Re-)Enable an uncertainty source so that it counts towards calculating the total uncertainty.

        :param error_name: error name
        :type error_name: str
        """
        _err_dict = self._get_error_by_name_raise(error_name)
        _err_dict['enabled'] = True
        self._on_error_change()

    def get_matching_errors(self, matching_criteria=None, matching_type='equal'):
        """Return a list of uncertainty objects fulfilling the specified matching criteria.

        Valid keys for **matching_criteria**:
            * ``name`` (the unique error name)
            * ``type`` (either ``simple`` or ``matrix``)
            * ``correlated`` (bool, only matches simple errors!)
            * ``relative`` (bool)

        .. note::
            The error objects contained in the dictionary are not copies, but the original error objects.
            Modifying them is possible, but not recommended.
            If you do modify any of them, the changes will not be reflected in the total error calculation until the
            error cache is cleared.
            This can be done by calling the private method :py:meth:`~_clear_total_error_cache`.

        :param matching_criteria: Key-value pairs specifying matching criteria. The resulting error array will only
                                  contain error objects matching *all* provided criteria.
                                  If :py:obj:`None`, all error objects are returned.
        :type matching_criteria: dict or None
        :param matching_type: How to perform the matching.
                              If ``'equal'``, the value in **matching_criteria** is checked for equality against the
                              stored value.
                              If ``'regex'``, the value in **matching_criteria** is interpreted as a regular expression
                              and is matched against the stored value.
        :type matching_type: str
        :return: Dict mapping error name to :py:obj:`~kafe2.core.error.GaussianErrorBase`-derived error objects.
        :rtype: dict[str, kafe2.core.error.GaussianErrorBase]
        """
        _result = copy(self._error_dicts)

        if matching_criteria is None:
            matching_criteria = dict()

        if matching_type == 'regex':
            raise NotImplementedError("Matching type 'regex' not yet implemented!")
        if matching_type != 'equal':
            raise NotImplementedError("Unknown matching type: '{}'! "
                                      "Available: ['equals']".format(matching_type))

        for _crit_key, _crit_value in six.iteritems(matching_criteria):
            # go through all errors, removing those that don't match
            for _error_name, _error_dict in list(_result.items()):  # do not use an iterator!
                if _crit_key == 'name' and _error_name == _crit_value:
                    continue
                elif _crit_key == 'type':
                    # type 'simple'
                    _err_obj = _error_dict['err']
                    if ((_crit_value == 'simple' and isinstance(_err_obj, SimpleGaussianError)) or
                            (_crit_value == 'matrix' and isinstance(_err_obj, MatrixGaussianError))):
                        continue
                elif _crit_key == 'correlated':
                    _err_obj = _error_dict['err']
                    try:
                        if _crit_value == bool(_err_obj.corr_coeff != 0):
                            continue
                    except AttributeError:
                        pass  # error is a MatrixGaussianError and will not be matched
                elif _crit_key == 'relative':
                    _err_obj = _error_dict['err']
                    if _err_obj.relative == _crit_value:
                        continue
                else:
                    # check the error dict keys for a match
                    _cmp_val = _error_dict.get(_crit_key, None)
                    if _cmp_val == _crit_value and _cmp_val is not None:
                        continue

                # no if clause above executed -> no match, remove error element
                _result.pop(_error_name)

                # exit loop if no remaining errors to examine
                if not _result:
                    break

            # exit loop if no remaining errors to examine
            if not _result:
                break

        return {_name: _entry['err'] for _name, _entry in six.iteritems(_result)}

    def get_error(self, error_name):
        """Return the uncertainty object holding the uncertainty.

        .. note::
            If you modify this object, the changes will not be reflected in the total error calculation until the error
            cache is cleared. This can be forced by calling :py:meth:`~enable_error`.

        :param error_name: error name
        :type error_name: str
        :return: error object
        :rtype: kafe2.core.error.GaussianErrorBase
        """
        return self._get_error_by_name_raise(error_name)

    def get_total_error(self):
        """Get the error object representing the total uncertainty.

        :return: error object representing the total uncertainty
        :rtype: kafe2.core.error.MatrixGaussianError
        """
        if self._total_error is None:
            self._calculate_total_error()
        return self._total_error



from copy import copy
import six
from abc import ABCMeta, abstractmethod
import numpy as np
import numdifftools as nd
from scipy.optimize import brentq

from ..error import CovMat


@six.add_metaclass(ABCMeta)
class MinimizerBase(object):

    ERRORDEF_CHI2 = 1.0
    ERRORDEF_NLL = 0.5

    def __init__(
            self, parameter_names, parameter_values, parameter_errors, function_to_minimize,
            tolerance=1e-6, errordef=ERRORDEF_CHI2):
        """
        :param parameter_names: the names of the parameters to vary during minimization.
        :type parameter_names: iterable of str
        :param parameter_values: the initial values of the parameters.
        :type parameter_values: iterable of float
        :param parameter_errors: the initial parameter errors, used to determine step size during
        minimization.
        :type parameter_errors: iterable of float
        :param function_to_minimize: the cost function to minimize.
        :type function_to_minimize: callable that returns a float
        :param tolerance: tolerance for convergence. Specifics depend on the used backend.
        :type tolerance: float
        :param errordef: difference in cost equivalent to the standard error.
        :type errordef: float
        """
        assert len(parameter_names) == len(parameter_values) == len(parameter_errors)
        self._invalidate_cache()  # initializes caches with None
        self.errordef = errordef
        self.tolerance = tolerance
        self._func_handle = function_to_minimize
        self._par_names = list(parameter_names)
        self.parameter_values = parameter_values
        self.parameter_errors = parameter_errors
        self._save_state_dict = dict()
        self._did_fit = False
        self._printed_inf_cost_warning = False

    def _invalidate_cache(self):
        """
        Invalidate the cache outside the used backend. Called at the end of `minimize` or `reset`.
        Parameter errors are **not** cleared because they need to be consitent.
        """
        self._fval = None
        self._par_asymm_err = None
        self._hessian = None
        self._hessian_inv = None
        self._par_cov_mat = None
        self._par_cor_mat = None

    def _save_state(self):
        """
        Save the state of this object including the state of the used backend, if any.
        """
        if self._par_asymm_err is None:
            self._save_state_dict['asymmetric_parameter_error'] = self._par_asymm_err
        else:
            self._save_state_dict['asymmetric_parameter_error'] = np.array(self._par_asymm_err)
        if self._hessian is None:
            self._save_state_dict['hessian'] = self._hessian
        else:
            self._save_state_dict['hessian'] = np.array(self._hessian)
        if self._hessian_inv is None:
            self._save_state_dict['hessian_inv'] = self._hessian_inv
        else:
            self._save_state_dict['hessian_inv'] = np.array(self._hessian_inv)
        if self._par_cov_mat is None:
            self._save_state_dict['par_cov_mat'] = self._par_cov_mat
        else:
            self._save_state_dict['par_cov_mat'] = np.array(self._par_cov_mat)
        if self._par_cor_mat is None:
            self._save_state_dict['par_cor_mat'] = self._par_cor_mat
        else:
            self._save_state_dict['par_cor_mat'] = np.array(self._par_cor_mat)
        self._save_state_dict['did_fit'] = self._did_fit

    def _load_state(self):
        """
        Load the state of this object including the state of the used backend, if any.
        """
        self._par_asymm_err = self._save_state_dict['asymmetric_parameter_error']
        if self._par_asymm_err is not None:
            self._par_asymm_err = np.array(self._par_asymm_err)
        self._hessian = self._save_state_dict['hessian']
        if self._hessian is not None:
            self._hessian = np.array(self._hessian)
        self._hessian_inv = self._save_state_dict['hessian_inv']
        if self._hessian_inv is not None:
            self._hessian_inv = np.array(self._hessian_inv)
        self._par_cov_mat = self._save_state_dict['par_cov_mat']
        if self._par_cov_mat is not None:
            self._par_cov_mat = np.array(self._par_cov_mat)
        self._par_cor_mat = self._save_state_dict['par_cor_mat']
        if self._par_cor_mat is not None:
            self._par_cor_mat = np.array(self._par_cor_mat)
        # Write back parameter values to nexus parameter nodes:
        self._func_wrapper_unpack_args(self.parameter_values)
        self._did_fit = self._save_state_dict['did_fit']

    def _func_wrapper(self, *args):
        """
        Wrapper for the cost function.
        :param args: cost function arguments (usually just the model function arguments) as varargs.
        :return: the cost function value.
        :rtype: float
        """
        _fval = self._func_handle(*args)
        if not self._printed_inf_cost_warning and np.isinf(_fval):
            print('Warning: the cost function has been evaluated as infinite. '
                  'The fit might not converge correctly.')
            self._printed_inf_cost_warning = True
        return _fval

    def _func_wrapper_unpack_args(self, args):
        """
        Wrapper for the cost function.
        :param args: cost function arguments (usually just the model function arguments) as an
        iterable.
        :type args: iterable of float
        :return: the cost function value.
        :rtype: float
        """
        return self._func_wrapper(*args)

    def _calculate_asymmetric_parameter_errors(self):  # TODO max calls
        """
        Calculate the asymmetric parameter errors. Works independently of the used backend, but
        might be overridden by a backend-specific implementation. Cost function must be negative log
        likelihood or chi squared to produce meaningful results. If no fit has been performed, all
        values in the return array are nan.
        :return: the asymmetric parameter errors for all parameters.
        :rtype numpy.ndarray of shape (num_pars, 2)
        """
        self.minimize()
        _ = self.parameter_errors  # call par error property so they're initialized for _save_state
        self._save_state()
        _asymm_par_errs = np.zeros(shape=self.parameter_values.shape + (2,))
        for _par_index, _par_name in enumerate(self.parameter_names):
            if self.is_fixed(_par_name):
                _asymm_par_errs[_par_index, :] = 0
            else:
                _target_chi_2 = self.function_value + 1.0
                _min_parameters = self.parameter_values

                _par_min = self.parameter_values[_par_index]
                _par_err = self.parameter_errors[_par_index]

                _cut_dn = self._find_cost_cut(
                    _par_name, _par_min - 2 * _par_err, _par_min, _target_chi_2, _min_parameters)
                _asymm_par_errs[_par_index, 0] = _cut_dn - _par_min

                _cut_up = self._find_cost_cut(
                    _par_name, _par_min, _par_min + 2 * _par_err, _target_chi_2, _min_parameters)
                _asymm_par_errs[_par_index, 1] = _cut_up - _par_min
                self._load_state()
        return _asymm_par_errs

    def _find_cost_cut(self, parameter_name, low, high, target_cost, min_parameters):
        """
        Utility function that finds the parameter value for a single parameter at which the cost
        function reaches a given value. The other parameters are **not** fixed. Instead the profile
        likelihood method is used.
        :param parameter_name: the name of the parameter to vary.
        :param low: lower bound of the parameter.
        :param high: upper bound of the parameter.
        :param target_cost: cost function value to find the cut for.
        :param min_parameters: parameter values at the cost function minimum.
        :return: the parameter value where the cost function value has the given value.
        :rtype: float
        """
        _all_pars_would_be_fixed = True
        for _par_name in self._par_names:
            if _par_name != parameter_name and not self.is_fixed(_par_name):
                _all_pars_would_be_fixed = False
                break

        def _profile(parameter_value):
            self.set_several(self.parameter_names, min_parameters)
            self.set(parameter_name, parameter_value)
            self._fval = None  # Clear fval cache
            if not _all_pars_would_be_fixed:
                self.fix(parameter_name)
                self.minimize()
                self.release(parameter_name)
            return self.function_value - target_cost

        return brentq(f=_profile, a=low, b=high, xtol=self.tolerance)

    def _remove_zeroes_for_fixed(self, matrix):
        """
        Takes a full error matrix and removes the rows and
        columns that corresponding to fixed parameters.
        """
        assert (matrix.shape == (self.num_pars, self.num_pars))

        _fixed_par_indices = [_i for _i, _par_name_i in enumerate(self._par_names)
                              if self.is_fixed(_par_name_i)]
        _submat = np.delete(
            np.delete(matrix, _fixed_par_indices, axis=0), _fixed_par_indices, axis=1)

        return _submat

    def _fill_in_zeroes_for_fixed(self, submatrix):
        """
        Takes the partial error matrix (submatrix) and adds
        rows and columns with 0.0 where the fixed
        parameters should go.
        """
        _mat = submatrix

        _fixed_par_indices = [_i for _i, _par_name_i in enumerate(self._par_names)
                              if self.is_fixed(_par_name_i)]
        for _id in _fixed_par_indices:
            _mat = np.insert(np.insert(_mat, _id, 0., axis=0), _id, 0., axis=1)

        assert (_mat.shape == (self.num_pars, self.num_pars))

        return _mat

    @property
    def did_fit(self):
        """
        :return: True if a fit has been performed after relevant input data has changed.
        """
        return self._did_fit

    @property
    def errordef(self):
        """
        :return: Difference in cost equivalent to the standard error.
        :rtype: float
        """
        return self._err_def

    @errordef.setter
    def errordef(self, err_def):
        assert err_def > 0
        self._err_def = err_def
        self.reset()

    @property
    def function_to_minimize(self):
        """
        :return: the cost function to minimize.
        :rtype: callable that returns a float
        """
        return self._func_handle

    @property
    def function_value(self):
        """
        :return: the cost function value using the current parameter values.
        :rtype: float
        """
        if self._fval is None:
            self._fval = self._func_handle(*self.parameter_values)
        if not self._printed_inf_cost_warning and np.isinf(self._fval):
            print('Warning: the cost function has been evaluated as infinite. '
                  'The fit might not converge correctly.')
            self._printed_inf_cost_warning = True
        return self._fval

    @property
    def num_pars(self):
        """
        :return: number of parameters to minimize the cost function value for.
        :rtype: int
        """
        return len(self._par_names)

    @property
    @abstractmethod
    def parameter_values(self):
        """
        :return: the current parameter values.
        :rtype: numpy.ndarray of shape (num_pars,)
        """

    @parameter_values.setter
    @abstractmethod
    def parameter_values(self, new_values):
        pass

    @property
    @abstractmethod
    def parameter_errors(self):
        """
        :return: the current parabolic parameter errors.
        :rtype: numpy.ndarray of shape (num_pars,)
        """

    @parameter_errors.setter
    @abstractmethod
    def parameter_errors(self, new_errors):
        pass

    @property
    def asymmetric_parameter_errors(self):
        """
        :return: the current asymmetric parameter errors derived from the likelihood.
        :rtype: numpy.ndarray of shape (num_pars, 2)
        """
        if not self.did_fit:
            return None
        if self._par_asymm_err is None:
            self._par_asymm_err = self._calculate_asymmetric_parameter_errors()
        if self._par_asymm_err is None:
            return None
        return self._par_asymm_err.copy()

    @property
    def asymmetric_parameter_errors_if_calculated(self):
        """
        As asymmetric_parameter_errors, but return None if the asymmetric parameter errors have not
        been calculated so far.
        :return: the current asymmetric parameter errors derived from the likelihood or None.
        :rtype: numpy.ndarray of shape (num_pars, 2) or None
        """
        return None if self._par_asymm_err is None else self._par_asymm_err.copy()

    @property
    def parameter_names(self):
        """
        :return: the names of the parameters.
        :rtype: list of str
        """
        return copy(self._par_names)

    @property
    def tolerance(self):
        """
        :return: tolerance to use during minimization. Details depend on the used backend.
        :rtype: float
        """
        return self._tol

    @tolerance.setter
    def tolerance(self, tolerance):
        assert tolerance > 0
        self._tol = tolerance
        self.reset()

    @property
    def hessian(self):
        """
        :return: the Hessian matrix calculated at the current parameter values.
        :rtype: numpy.ndarray of shape (num_pars, num_pars)
        """
        if not self.did_fit:
            return None
        if self._hessian is None:
            self._hessian = nd.Hessian(self._func_wrapper_unpack_args)(self.parameter_values)
            assert(np.all(self._hessian == self._hessian.T))
            # Write back parameter values to nexus parameter nodes:
            self._func_wrapper_unpack_args(self.parameter_values)
        return self._hessian.copy()

    @property
    def hessian_inv(self):
        """
        :return: the inverse of the Hessian matrix calculated at the current parameter values.
        :rtype: numpy.ndarray of shape (num_pars, num_pars)
        """
        if not self.did_fit:
            return None
        if self._hessian_inv is None:
            _hessian = self.hessian  # including zeroes for fixed parameter
            _subhessian = self._remove_zeroes_for_fixed(_hessian)
            _subhessian_inv = np.linalg.inv(_subhessian)
            self._hessian_inv = self._fill_in_zeroes_for_fixed(_subhessian_inv)
            # ensure symmetric
            self._hessian_inv = 0.5 * (self._hessian_inv + self._hessian_inv.T)
            assert (np.all(self._hessian_inv == self._hessian_inv.T))
        return self._hessian_inv.copy()

    @property
    def cov_mat(self):
        """
        :return: the parameter covariance matrix calculated at the current parameter values.
        :rtype: numpy.ndarray of shape (num_pars, num_pars)
        """
        if not self.did_fit:
            return None
        if self._par_cov_mat is None:
            self._par_cov_mat = self.hessian_inv * 2.0 * self.errordef
        return self._par_cov_mat.copy()

    @property
    def cor_mat(self):
        """
        :return: the parameter correlation matrix calculated at the current parameter values.
        :rtype: numpy.ndarray of shape (num_pars, num_pars)
        """
        if not self.did_fit:
            return None
        if self._par_cor_mat is None:
            _cov_mat = self.cov_mat
            if _cov_mat is None:
                return None
            _subcov_mat = self._remove_zeroes_for_fixed(_cov_mat)
            _subcor_mat = CovMat(_subcov_mat).cor_mat
            self._par_cor_mat = self._fill_in_zeroes_for_fixed(_subcor_mat)
        return self._par_cor_mat.copy()

    def reset(self):
        """Clears caches and resets the internal state of the used backend (if any)."""
        self._invalidate_cache()
        self._did_fit = False

    @abstractmethod
    def set(self, parameter_name, parameter_value):
        """
        Sets one parameter to the given value.
        :param parameter_name: the name of the parameter.
        :type parameter_name: str
        :param parameter_value: the value to set the parameter to.
        :type parameter_value: float
        """

    def set_several(self, parameter_names, parameter_values):
        """
        Sets multiple parameters to the given values.
        :param parameter_names: the names of the parameter.
        :type parameter_names: iterable of str
        :param parameter_values: the values to set the parameters to.
        :type parameter_values: iterable of float
        """
        for _pn, _pv in zip(parameter_names, parameter_values):
            self.set(_pn, _pv)

    @abstractmethod
    def limit(self, parameter_name, parameter_bounds):
        """
        Limit a parameter value to a given closed interval during minimization.
        :param parameter_name: the name of the parameter to limit.
        :type parameter_name: str
        :param parameter_bounds: the lower and upper bounds of the parameter.
        :type parameter_bounds: tuple consisting of two floats: (lower_bound, upper_bound)
        """

    @abstractmethod
    def unlimit(self, parameter_name):
        """
        Reverse the limitation of a parameter.
        :param parameter_name: the name of the parameter to unlimit.
        :type parameter_name: str
        """

    @abstractmethod
    def fix(self, parameter_name):
        """
        Fix a parameter to its current value. It will be treated as constant during minimization.
        :param parameter_name: the name of the parameter to fix.
        :type parameter_name: str
        """

    def fix_several(self, parameter_names):
        """
        Fix several parameters to their current values. They will be treated as constant during
        minimization.
        :param parameter_names: the names of the parameters to fix.
        :type parameter_names: iterable of str
        """
        for _pn in parameter_names:
            self.fix(_pn)

    @abstractmethod
    def is_fixed(self, parameter_name):
        """
        Check if the given parameter is currently fixed.
        :param parameter_name: the name of the parameter to check fixed status for.
        :type parameter_name: str
        :return: True if the parameter is currently fixed.
        :rtype: bool
        """

    @abstractmethod
    def release(self, parameter_name):
        """
        Revert the fixation of a parameter. It will no longer be treated as constant during
        minimization.
        :param parameter_name: the name of the parameter to release.
        :type parameter_name: str
        """

    def release_several(self, parameter_names):
        """
        Revert the fixation of several parameters. They will no longer be treated as constant during
        minimization.
        :param parameter_names: the names of the parameter to release.
        :type parameter_names: iterable of str
        """
        for _pn in parameter_names:
            self.release(_pn)

    @abstractmethod
    def minimize(self):  # TODO max calls
        """
        Minimize the cost function with the used backend.
        """

    @abstractmethod
    def contour(self, parameter_name_1, parameter_name_2, sigma=1.0, **minimizer_contour_kwargs):
        """
        Calculate a 2D contour using the profile likelihood method: two parameters are fixed while
        the rest are varied to minimize the cost function. The parameter values of the fixed
        parameters at which the cost function value has a given offset relative to the cost function
        minimum constitute the contour.
        :param parameter_name_1: the name of the first parameter to fix.
        :type parameter_name_1: str
        :param parameter_name_2:
        :type parameter_name_2: str
        :param sigma: difference between the cost function minimum and the contour.
        :type sigma: float
        :param minimizer_contour_kwargs: backend-specific kwargs.
        :return: the calculated contour
        :rtype: kafe2.core.contour.Contour
        """

    @abstractmethod
    def profile(self, parameter_name, bins=20, bound=2, subtract_min=False):
        """
        Calculate a 1D profile using the profile likelihood method: a single parameter is fixed
        while the rest are varied to minimize the cost function. The mapping of parameter value to
        cost function value around the minimum is the profile.
        The interval [par_min - par_err * bound, par_min + par_err * bound] is profiled by this
        method.
        :param parameter_name: the name of the parameter to fix.
        :type parameter_name: str
        :param bins: the number of points to evaluate the cost function at.
        :type bins: int
        :param bound: parameter determining the size of the interval to profile (see above).
        :type bound: float
        :param subtract_min: if True, subtract the cost function value of the minimum from the cost
        function values of the profile.
        :type subtract_min: bool
        :return: the parameter values of the fixed parameter and the corresponding cost function
        values.
        :rtype: numpy.ndarray of shape (2, bins)
        """
